from __future__ import annotations

import importlib
import io
import json
from pathlib import Path
from typing import Any, BinaryIO, Sequence
from zipfile import ZipFile

import skops

from ._audit import NODE_TYPE_MAPPING, audit_tree, get_tree
from ._utils import LoadContext, SaveContext, _get_state, get_state

# We load the dispatch functions from the corresponding modules and register
# them. Old protocols are found in the 'old/' directory, with the protocol
# version appended to the corresponding module name.
modules = ["._general", "._numpy", "._scipy", "._sklearn"]
modules.extend([".old._general_v0", ".old._numpy_v0"])
for module_name in modules:
    # register exposed functions for get_state and get_tree
    module = importlib.import_module(module_name, package="skops.io")
    for cls, method in getattr(module, "GET_STATE_DISPATCH_FUNCTIONS", []):
        _get_state.register(cls)(method)
    # populate the the dict used for dispatching get_tree functions
    NODE_TYPE_MAPPING.update(module.NODE_TYPE_MAPPING)


def _save(obj: Any) -> io.BytesIO:
    buffer = io.BytesIO()

    with ZipFile(buffer, "w") as zip_file:
        save_context = SaveContext(zip_file=zip_file)
        state = get_state(obj, save_context)
        save_context.clear_memo()

        state["protocol"] = save_context.protocol
        state["_skops_version"] = skops.__version__
        zip_file.writestr("schema.json", json.dumps(state, indent=2))

    return buffer


def dump(obj: Any, file: str | Path | BinaryIO) -> None:
    """Save an object using the skops persistence format.

    Skops aims at providing a secure persistence feature that does not rely on
    :mod:`pickle`, which is inherently insecure. For more information, please
    visit the :ref:`persistence` documentation.

    .. warning::

        This feature is heavily under development, which means the API is
        unstable and there might be security issues at the moment. Therefore,
        use caution when loading files from sources you don't trust.

    Parameters
    ----------
    obj: object
        The object to be saved. Usually a scikit-learn compatible model.

    file: str, path, or file-like object
        The file name. A zip archive will automatically created. As a matter of
        convention, we recommend to use the ".skops" file extension, e.g.
        ``save(model, "my-model.skops")``.

    """
    buffer = _save(obj)

    if isinstance(file, (str, Path)):
        with open(file, "wb") as f:
            f.write(buffer.getbuffer())
    else:
        file.write(buffer.getbuffer())


def dumps(obj: Any) -> bytes:
    """Save an object using the skops persistence format as a bytes object.

    .. warning::

        This feature is heavily under development, which means the API is
        unstable and there might be security issues at the moment. Therefore,
        use caution when loading files from sources you don't trust.

    Parameters
    ----------
    obj: object
        The object to be saved. Usually a scikit-learn compatible model.

    """
    buffer = _save(obj)
    return buffer.getbuffer().tobytes()


def load(file: str | Path, trusted: bool | Sequence[str] = False) -> Any:
    """Load an object saved with the skops persistence format.

    Skops aims at providing a secure persistence feature that does not rely on
    :mod:`pickle`, which is inherently insecure. For more information, please
    visit the :ref:`persistence` documentation.

    .. warning::

        This feature is heavily under development, which means the API is
        unstable and there might be security issues at the moment. Therefore,
        use caution when loading files from sources you don't trust.

    Parameters
    ----------
    file: str or pathlib.Path
        The file name of the object to be loaded.

    trusted: bool, or list of str, default=False
        If ``True``, the object will be loaded without any security checks. If
        ``False``, the object will be loaded only if there are only trusted
        objects in the dumped file. If a list of strings, the object will be
        loaded only if there are only trusted objects and objects of types
        listed in ``trusted`` are in the dumped file.

    Returns
    -------
    instance: object
        The loaded object.

    """
    with ZipFile(file, "r") as input_zip:
        schema = json.loads(input_zip.read("schema.json"))
        load_context = LoadContext(src=input_zip, protocol=schema["protocol"])
        tree = get_tree(schema, load_context)
        audit_tree(tree, trusted)
        instance = tree.construct()

    return instance


def loads(data: bytes, trusted: bool | Sequence[str] = False) -> Any:
    """Load an object saved with the skops persistence format from a bytes
    object.

    .. warning::

        This feature is heavily under development, which means the API is
        unstable and there might be security issues at the moment. Therefore,
        use caution when loading files from sources you don't trust.

    Parameters
    ----------
    data: bytes
        The dumped data to be loaded in bytes format.

    trusted: bool, or list of str, default=False
        If ``True``, the object will be loaded without any security checks. If
        ``False``, the object will be loaded only if there are only trusted
        objects in the dumped file. If a list of strings, the object will be
        loaded only if there are only trusted objects and objects of types
        listed in ``trusted`` are in the dumped file.

    Returns
    -------
    instance: object
        The loaded object.
    """
    if isinstance(data, str):
        raise TypeError("Can't load skops format from string, pass bytes")

    with ZipFile(io.BytesIO(data), "r") as zip_file:
        schema = json.loads(zip_file.read("schema.json"))
        load_context = LoadContext(src=zip_file, protocol=schema["protocol"])
        tree = get_tree(schema, load_context)
        audit_tree(tree, trusted)
        instance = tree.construct()

    return instance


def get_untrusted_types(
    *, data: bytes | None = None, file: str | Path | None = None
) -> list[str]:
    """Get a list of untrusted types in a skops dump.

    Parameters
    ----------
    data: bytes
        The data to be checked, in bytes format.

    file: str or Path
        The file to be checked.

    Returns
    -------
    untrusted_types: list of str
        The list of untrusted types in the dump.

    Notes
    -----
    Only one of data or file should be passed.
    """
    if data and file:
        raise ValueError("Only one of data or file should be passed.")
    if not data and not file:
        raise ValueError("Exactly one of data or file should be passed.")

    content: io.BytesIO | str | Path
    if data:
        content = io.BytesIO(data)
    else:
        # mypy doesn't understand that file cannot be None here, thus ignore
        content = file  # type: ignore

    with ZipFile(content, "r") as zip_file:
        schema = json.loads(zip_file.read("schema.json"))
        tree = get_tree(
            schema, load_context=LoadContext(src=zip_file, protocol=schema["protocol"])
        )
        untrusted_types = tree.get_unsafe_set()

    return sorted(untrusted_types)



import logging
import pathlib
import pickle
from unittest import mock

import numpy as np
import pytest

from skops.cli import _convert
from skops.io import load


class MockUnsafeType:
    def __init__(self):
        pass


class TestConvert:
    model_name = "some_model_name"

    @pytest.fixture
    def safe_obj(self):
        return np.ndarray([1, 2, 3, 4])

    @pytest.fixture
    def unsafe_obj(self):
        return MockUnsafeType()

    @pytest.fixture
    def pkl_path(self, tmp_path):
        return tmp_path / f"{self.model_name}.pkl"

    @pytest.fixture
    def skops_path(self, tmp_path):
        return tmp_path / f"{self.model_name}.skops"

    @pytest.fixture
    def write_safe_file(self, pkl_path, safe_obj):
        with open(pkl_path, "wb") as f:
            pickle.dump(safe_obj, f)

    @pytest.fixture
    def write_unsafe_file(self, pkl_path, unsafe_obj):
        with open(pkl_path, "wb") as f:
            pickle.dump(unsafe_obj, f)

    def test_base_case_works_as_expected(
        self, pkl_path, tmp_path, skops_path, write_safe_file, safe_obj, caplog
    ):
        mock_logger = mock.MagicMock()
        _convert._convert_file(pkl_path, skops_path, logger=mock_logger)
        persisted_obj = load(skops_path)
        assert np.array_equal(persisted_obj, safe_obj)

        # Check no warnings or errors raised
        mock_logger.warning.assert_not_called()
        mock_logger.error.assert_not_called()

    def test_unsafe_case_works_as_expected(
        self, pkl_path, tmp_path, skops_path, write_unsafe_file, caplog
    ):
        caplog.set_level(logging.WARNING)
        _convert._convert_file(pkl_path, skops_path)
        persisted_obj = load(skops_path, trusted=True)

        assert isinstance(persisted_obj, MockUnsafeType)

        # check logging has warned that an unsafe type was found
        assert MockUnsafeType.__name__ in caplog.text


class TestMain:
    @staticmethod
    def assert_called_correctly(
        mock_convert: mock.MagicMock,
        path,
        output_file=None,
    ):
        if not output_file:
            output_file = pathlib.Path.cwd() / f"{pathlib.Path(path).stem}.skops"
        mock_convert.assert_called_once_with(input_file=path, output_file=output_file)

    @mock.patch("skops.cli._convert._convert_file")
    def test_base_works_as_expected(self, mock_convert: mock.MagicMock):
        path = "123.pkl"
        namespace, _ = _convert.format_parser().parse_known_args([path])

        _convert.main(namespace)
        self.assert_called_correctly(mock_convert, path)

    @mock.patch("skops.cli._convert._convert_file")
    @pytest.mark.parametrize(
        "input_path, output_file, expected_path",
        [
            ("abc.123", "some/file/path.out", "some/file/path.out"),
            ("abc.123", None, pathlib.Path.cwd() / "abc.skops"),
        ],
        ids=["Given an output path", "No output path"],
    )
    def test_with_output_dir_works_as_expected(
        self, mock_convert: mock.MagicMock, input_path, output_file, expected_path
    ):
        if output_file is not None:
            args = [input_path, "--output", output_file]
        else:
            args = [input_path]

        namespace, _ = _convert.format_parser().parse_known_args(args)

        _convert.main(namespace)
        self.assert_called_correctly(
            mock_convert, path=input_path, output_file=expected_path
        )

    @mock.patch("skops.cli._convert._convert_file")
    @pytest.mark.parametrize(
        "verbosity, expected_level",
        [
            ("", logging.WARNING),
            ("-v", logging.INFO),
            ("--verbose", logging.INFO),
            ("-vv", logging.DEBUG),
            ("-v -v", logging.DEBUG),
            ("-vvvvv", logging.DEBUG),
            ("--verbose --verbose", logging.DEBUG),
        ],
    )
    def test_given_log_levels_works_as_expected(
        self, mock_convert: mock.MagicMock, verbosity, expected_level, caplog
    ):
        input_path = "abc.def"
        output_path = "bde.skops"
        args = [input_path, "--output", output_path, verbosity.split()]

        namespace, _ = _convert.format_parser().parse_known_args(args)

        _convert.main(namespace)
        self.assert_called_correctly(
            mock_convert, path=input_path, output_file=output_path
        )

        assert caplog.at_level(expected_level)



import io
import json
import re
from contextlib import suppress
from zipfile import ZipFile

import numpy as np
import pytest
from sklearn.linear_model import LogisticRegression
from sklearn.pipeline import FeatureUnion, Pipeline
from sklearn.preprocessing import FunctionTransformer, StandardScaler

from skops.io import dumps, get_untrusted_types
from skops.io._audit import Node, audit_tree, check_type, get_tree, temp_setattr
from skops.io._general import DictNode, JsonNode, ObjectNode, dict_get_state
from skops.io._utils import LoadContext, SaveContext, get_state, gettype


class CustomType:
    """A custom untrusted class."""

    def __init__(self, value):
        self.value = value


@pytest.mark.parametrize(
    "module_name, type_name, trusted, expected",
    [
        ("sklearn", "Pipeline", ["sklearn.Pipeline"], True),
        ("sklearn", "Pipeline", ["sklearn.preprocessing.StandardScaler"], False),
        ("sklearn", "Pipeline", True, True),
        ("builtins", "int", ["builtins.int"], True),
        ("builtins", "int", [], False),
    ],
    ids=["list-True", "list-False", "True", "int-True", "int-False"],
)
def test_check_type(module_name, type_name, trusted, expected):
    assert check_type(module_name, type_name, trusted) == expected


def test_audit_tree_untrusted():
    var = {"a": CustomType(1), 2: CustomType(2)}
    state = dict_get_state(var, SaveContext(None, 0, {}))
    node = DictNode(state, LoadContext(None, -1), trusted=False)
    with pytest.raises(
        TypeError,
        match=re.escape(
            "Untrusted types found in the file: ['test_audit.CustomType']."
        ),
    ):
        audit_tree(node, trusted=False)

    # there shouldn't be an error with trusted=True
    audit_tree(node, trusted=True)

    untrusted_list = get_untrusted_types(data=dumps(var))
    assert untrusted_list == ["test_audit.CustomType"]

    # passing the type would fix it.
    audit_tree(node, trusted=untrusted_list)


def test_audit_tree_defaults():
    # test that the default types are trusted
    var = {"a": 1, 2: "b"}
    state = dict_get_state(var, SaveContext(None, 0, {}))
    node = DictNode(state, LoadContext(None, -1), trusted=False)
    audit_tree(node, trusted=[])


@pytest.mark.parametrize(
    "trusted, defaults, expected",
    [
        (True, None, True),
        (False, int, ["builtins.int"]),
        ([int], None, ["builtins.int"]),
    ],
    ids=["trusted", "untrusted", "untrusted_list"],
)
def test_Node_get_trusted(trusted, defaults, expected):
    assert Node._get_trusted(trusted, defaults) == expected


@pytest.mark.parametrize(
    "values, is_safe",
    [
        ([1, 2], True),
        ([1, {1: 2}], True),
        ([1, {1: CustomType(1)}], False),
        (eval, False),
        (pytest.mark.parametrize, False),
    ],
    ids=["int", "dict", "untrusted", "eval", "parametrize"],
)
def test_list_safety(values, is_safe):
    content = dumps(values)

    with ZipFile(io.BytesIO(content), "r") as zip_file:
        schema = json.loads(zip_file.read("schema.json"))
        tree = get_tree(schema, load_context=LoadContext(src=zip_file, protocol=-1))
        assert tree.is_safe() == is_safe


def test_gettype_error():
    msg = "Object None of module test is unknown"
    with pytest.raises(ValueError, match=msg):
        gettype(module_name="test", cls_or_func=None)

    msg = "Object test of module None is unknown"
    with pytest.raises(ValueError, match=msg):
        gettype(module_name=None, cls_or_func="test")

    # ImportError if the module cannot be imported
    with pytest.raises(ImportError):
        gettype(module_name="invalid-module", cls_or_func="invalid-type")


@pytest.mark.parametrize(
    "data, file, exception, message",
    [
        ("not-none", "not-none", ValueError, "Only one of data or file"),
        (None, None, ValueError, "Exactly one of data or file should be passed"),
        ("string", None, TypeError, "a bytes-like object is required, not 'str'"),
    ],
    ids=["both", "neither", "string-data"],
)
def test_get_untrusted_types_validation(data, file, exception, message):
    with pytest.raises(exception, match=message):
        get_untrusted_types(data=data, file=file)


def test_temp_setattr():
    # Test that temp_setattr works as expected
    class A:
        def __init__(self):
            self.a = 1

    temp = A()
    with suppress(ValueError):
        with temp_setattr(temp, a=2, b=3):
            assert temp.a == 2
            assert temp.b == 3
            raise ValueError  # to make sure context manager handles exceptions

    assert temp.a == 1
    assert not hasattr(temp, "b")


def test_complex_pipeline_untrusted_set():
    # fmt: off
    clf = Pipeline([
        ("features", FeatureUnion([
            ("scaler", StandardScaler()),
            ("sqrt", FunctionTransformer(
                    func=np.sqrt,
                    inverse_func=np.square,
                )),
        ])),
        ("clf", LogisticRegression(random_state=0, solver="liblinear")),
    ])
    # fmt: on

    untrusted = get_untrusted_types(data=dumps(clf))
    type_names = [x.split(".")[-1] for x in untrusted]
    assert type_names == ["sqrt", "square"]


def test_format_object_node():
    estimator = LogisticRegression(random_state=0, solver="liblinear")
    state = get_state(estimator, SaveContext(None))
    node = ObjectNode(state, LoadContext(None, -1))
    expected = "sklearn.linear_model._logistic.LogisticRegression"
    assert node.format() == expected


@pytest.mark.parametrize(
    "inp, expected",
    [
        ("hello", 'json-type("hello")'),
        (123, "json-type(123)"),
        (0.456, "json-type(0.456)"),
        (True, "json-type(true)"),
        (False, "json-type(false)"),
        (None, "json-type(null)"),
    ],
)
def test_format_json_node(inp, expected):
    state = get_state(inp, SaveContext(None))
    node = JsonNode(state, LoadContext(None, -1))
    assert node.format() == expected



import logging
import pathlib
import sys
from unittest import mock

import pytest

from skops.cli.entrypoint import main_cli


class TestEntrypoint:
    """Integration tests that check that entrypoint calls pass through correctly.
    Full coverage of individual entrypoint calls should be done in their own classes.
    """

    @pytest.fixture(autouse=True)
    def clear_argv(self):
        # Required to clear argv in case Pytest is called on this specific function.
        # Otherwise, clogs parser.parse_known_args() in argparse
        sys.argv = [""]

    @mock.patch("skops.cli._convert._convert_file")
    def test_convert_works_as_expected(
        self,
        convert_file_mock: mock.MagicMock,
        caplog,
    ):
        """
        Intended as a unit test to make sure,
        given 'convert' as the first argument,
        the parser is configured correctly
        """

        args = ["convert", "abc.def"]

        main_cli(args)
        convert_file_mock.assert_called_once_with(
            input_file="abc.def", output_file=pathlib.Path.cwd() / "abc.skops"
        )

        assert caplog.at_level(logging.WARNING)



import logging


def get_log_level(level: int = 0) -> int:
    """Takes in verbosity from a CLI entrypoint (number of times -v specified),
    and sets the logger to the required log level"""

    all_levels = [logging.WARNING, logging.INFO, logging.DEBUG]

    if level > len(all_levels):
        level = len(all_levels) - 1
    elif level < 0:
        level = 0

    return all_levels[level]



import builtins
from unittest.mock import patch

import pytest


@pytest.fixture
def pandas_not_installed():
    # patch import so that it raises an ImportError when trying to import
    # pandas. This works because pandas is only imported lazily.

    orig_import = builtins.__import__

    def mock_import(name, *args, **kwargs):
        if name == "pandas":
            raise ImportError
        return orig_import(name, *args, **kwargs)

    with patch("builtins.__import__", side_effect=mock_import):
        yield


@pytest.fixture
def matplotlib_not_installed():
    # patch import so that it raises an ImportError when trying to import
    # matplotlib. This works because matplotlib is only imported lazily.

    # ugly way of removing matplotlib from cached imports
    import sys

    for key in list(sys.modules.keys()):
        if key.startswith("matplotlib"):
            del sys.modules[key]

    orig_import = builtins.__import__

    def mock_import(name, *args, **kwargs):
        if name == "matplotlib":
            raise ImportError
        return orig_import(name, *args, **kwargs)

    with patch("builtins.__import__", side_effect=mock_import):
        yield

    import matplotlib  # noqa


@pytest.fixture
def rich_not_installed():
    orig_import = builtins.__import__

    def mock_import(name, *args, **kwargs):
        if name == "rich":
            raise ImportError
        return orig_import(name, *args, **kwargs)

    with patch("builtins.__import__", side_effect=mock_import):
        yield



"""Templates for model cards

To add a new template, define it as a dictionary where the key is the section
and the value is the content of the section. If the content is empty but should
be filled by the user, set it to be the ``CONTENT_PLACEHOLDER``.

After defining the template itself, add it as another enum value in the
``Templates`` enum.

Finally, if there is a corresponding section in the new template, some methods
on the ``Card`` class should be adjusted to make use of the template. First of
all, ``_fill_default_sections`` should be used to populate the model card with
the template.

Furthermore, some specific methods might require adjusting. For example, the
``Card._add_hyperparams`` method will add a table of model hyperparameters, but
it needs to know in what section to put them. So if the template contains a
corresponding section, modify the method to put the hyperparameters inside that
section.

"""

from enum import Enum


class Templates(Enum):
    skops = "skops"


CONTENT_PLACEHOLDER = "[More Information Needed]"
"""When there is a section but content has yet to be added by the user, show
this"""

# fmt: off
SKOPS_TEMPLATE = {
    "Model description": CONTENT_PLACEHOLDER,
    "Model description/Intended uses & limitations": CONTENT_PLACEHOLDER,
    "Model description/Training Procedure": CONTENT_PLACEHOLDER,
    "Model description/Training Procedure/Hyperparameters": CONTENT_PLACEHOLDER,
    "Model description/Training Procedure/Model Plot": CONTENT_PLACEHOLDER,
    "Model description/Evaluation Results": CONTENT_PLACEHOLDER,
    "How to Get Started with the Model": CONTENT_PLACEHOLDER,
    "Model Card Authors": (
        f"This model card is written by following authors:\n\n{CONTENT_PLACEHOLDER}"
    ),
    "Model Card Contact": (
        "You can contact the model card authors through following channels:\n"
        f"{CONTENT_PLACEHOLDER}"
    ),
    "Citation": (
        "Below you can find information related to citation.\n\n**BibTeX:**\n```\n"
        f"{CONTENT_PLACEHOLDER}\n```"
    ),
}

# The template below corresponds to the HF Hub default template, but is geared
# towards deep learning models, especially language models, and thus is not a
# good fit for most sklearn models.
_HUB_TEMPLATE = {
    "Model Card": "",
    # Provide a quick summary of what the model is/does.
    "Model Details": "",
    "Model Details/Model Description": "",
    # Provide a longer summary of what this model is.
    "Model Details/Model Description/Developed by": CONTENT_PLACEHOLDER,
    "Model Details/Model Description/Shared by [optional]": CONTENT_PLACEHOLDER,
    "Model Details/Model Description/Model type": CONTENT_PLACEHOLDER,
    "Model Details/Model Description/Language(s) (NLP)": CONTENT_PLACEHOLDER,
    "Model Details/Model Description/License": CONTENT_PLACEHOLDER,
    "Model Details/Model Description/Finetuned from model [optional]": CONTENT_PLACEHOLDER,
    "Model Details/Model Description/Resources for more information": CONTENT_PLACEHOLDER,

    "Uses": "",
    # Address questions around how the model is intended to be used, including
    # the foreseeable users of the model and those affected by the model.
    "Uses/Direct Use": CONTENT_PLACEHOLDER,
    # This section is for the model use without fine-tuning or plugging into a
    # larger ecosystem/app.
    "Uses/Downstream Use [optional]": CONTENT_PLACEHOLDER,
    # This section is for the model use when fine-tuned for a task, or when
    # plugged into a larger ecosystem/app.
    "Uses/Out-of-Scope Use": CONTENT_PLACEHOLDER,
    # This section addresses misuse, malicious use, and uses that the model will
    # not work well for.

    "Bias, Risks, and Limitations": CONTENT_PLACEHOLDER,
    # This section is meant to convey both technical and sociotechnical
    # limitations.
    "Bias, Risks, and Limitations/Recommendations": (
        "Users (both direct and downstream) should be made aware of the risks, biases "
        "and limitations of the model. More information needed for further "
        "recommendations."
    ),
    # This section is meant to convey recommendations with respect to the bias,
    # risk, and technical limitations.

    "Training Details": "",
    "Training Details/Training Data": CONTENT_PLACEHOLDER,
    # This should link to a Data Card, perhaps with a short stub of information
    # on what the training data is all about as well as documentation related to
    # data pre-processing or additional filtering.
    "Training Details/Training Procedure [optional]": "",
    # This relates heavily to the Technical Specifications. Content here should
    # link to that section when it is relevant to the training procedure.
    "Training Details/Training Procedure [optional]/Preprocessing": CONTENT_PLACEHOLDER,
    "Training Details/Training Procedure [optional]/Speeds, Sizes, Times": CONTENT_PLACEHOLDER,
    # This section provides information about throughput, start/end time,
    # checkpoint size if relevant, etc.

    "Evaluation": "",
    # This section describes the evaluation protocols and provides the results.
    "Evaluation/Testing Data, Factors & Metrics": "",
    "Evaluation/Testing Data, Factors & Metrics/Testing Data": CONTENT_PLACEHOLDER,
    # This should link to a Data Card if possible
    "Evaluation/Testing Data, Factors & Metrics/Factors": CONTENT_PLACEHOLDER,
    # These are the things the evaluation is disaggregating by, e.g.,
    # subpopulations or domains.
    "Evaluation/Testing Data, Factors & Metrics/Metrics": CONTENT_PLACEHOLDER,
    # These are the evaluation metrics being used, ideally with a description of
    # why.
    "Evaluation/Results": CONTENT_PLACEHOLDER,

    "Model Examination [optional]": CONTENT_PLACEHOLDER,
    # Relevant interpretability work for the model goes here.

    "Environmental Impact": (
        "Carbon emissions can be estimated using the "
        "[Machine Learning Impact calculator](https://mlco2.github.io/impact#compute) "
        "presented in [Lacoste et al. (2019)](https://arxiv.org/abs/1910.09700)."
    ),
    # Total emissions (in grams of CO2eq) and additional considerations, such as
    # electricity usage, go here. Edit the suggested text below accordingly"
    "Environmental Impact/Hardware Type": CONTENT_PLACEHOLDER,
    "Environmental Impact/Hours used": CONTENT_PLACEHOLDER,
    "Environmental Impact/Cloud Provider": CONTENT_PLACEHOLDER,
    "Environmental Impact/Compute Region": CONTENT_PLACEHOLDER,
    "Environmental Impact/Carbon Emitted": CONTENT_PLACEHOLDER,

    "Technical Specifications [optional]": "",
    "Technical Specifications [optional]/Model Architecture and Objective": CONTENT_PLACEHOLDER,
    "Technical Specifications [optional]/Compute Infrastructure": CONTENT_PLACEHOLDER,
    "Technical Specifications [optional]/Compute Infrastructure/Hardware": CONTENT_PLACEHOLDER,
    "Technical Specifications [optional]/Compute Infrastructure/Software": CONTENT_PLACEHOLDER,

    "Citation [optional]": "",
    # If there is a paper or blog post introducing the model, the APA and Bibtex
    # information for that should go in this section.
    "Citation [optional]/BibTeX": CONTENT_PLACEHOLDER,
    "Citation [optional]/APA": CONTENT_PLACEHOLDER,

    "Glossary [optional]": "",
    # If relevant, include terms and calculations in this section that can help
    # readers understand the model or model card.

    "More Information [optional]": CONTENT_PLACEHOLDER,
    "Model Card Authors [optional]": CONTENT_PLACEHOLDER,
    "Model Card Contact": CONTENT_PLACEHOLDER,
    "How to Get Started with the Model": f"""Use the code below to get started with the model.

<details>
<summary> Click to expand </summary>

{CONTENT_PLACEHOLDER}

</details>""",
}
# fmt: on



import sys

# PEP0440 compatible formatted version, see:
# https://www.python.org/dev/peps/pep-0440/
#
# Generic release markers:
#   X.Y.0   # For first release after an increment in Y
#   X.Y.Z   # For bugfix releases
#
# Admissible pre-release markers:
#   X.Y.ZaN   # Alpha release
#   X.Y.ZbN   # Beta release
#   X.Y.ZrcN  # Release Candidate
#   X.Y.Z     # Final release
#
# Dev branch marker is: 'X.Y.dev' or 'X.Y.devN' where N is an integer.
# 'X.Y.dev0' is the canonical version of 'X.Y.dev'
#
__version__ = "0.6.0"

try:
    # This variable is injected in the __builtins__ by the build
    # process. It is used to enable importing subpackages of skops when
    # the binaries are not built
    # mypy error: Cannot determine type of '__SKOPS_SETUP__'
    __SKOPS_SETUP__  # type: ignore
except NameError:
    __SKOPS_SETUP__ = False

if __SKOPS_SETUP__:
    sys.stderr.write("Partial import of the library during the build process.\n")
    # We are not importing the rest of the library during the build
    # process, as it may not be compiled yet or cause immature import



class UnsupportedTypeException(TypeError):
    """Raise when an object of this type is known to be unsupported"""

    def __init__(self, obj):
        super().__init__(
            f"Objects of type {obj.__class__.__name__} are not supported yet."
        )


class UntrustedTypesFoundException(TypeError):
    """Raise when some untrusted objects are found in the file."""

    def __init__(self, unsafe):
        super().__init__(f"Untrusted types found in the file: {sorted(unsafe)}.")



from __future__ import annotations

import io
from typing import Any, Sequence

import numpy as np

from ._audit import Node, get_tree
from ._general import function_get_state
from ._protocol import PROTOCOL
from ._utils import LoadContext, SaveContext, get_module, get_state, gettype
from .exceptions import UnsupportedTypeException


def ndarray_get_state(obj: Any, save_context: SaveContext) -> dict[str, Any]:
    res = {
        "__class__": obj.__class__.__name__,
        "__module__": get_module(type(obj)),
        "__loader__": "NdArrayNode",
    }

    try:
        # If the dtype is object, np.save should not work with
        # allow_pickle=False, therefore we convert them to a list and
        # recursively call get_state on it.
        if obj.dtype == object:
            obj_serialized = get_state(obj.tolist(), save_context)
            res["content"] = obj_serialized["content"]
            res["type"] = "json"
            res["shape"] = get_state(obj.shape, save_context)
        else:
            data_buffer = io.BytesIO()
            np.save(data_buffer, obj, allow_pickle=False)
            # Memoize the object and then check if it's file name (containing
            # the object id) already exists. If it does, there is no need to
            # save the object again. Memoizitation is necessary since for
            # ephemeral objects, the same id might otherwise be reused.
            obj_id = save_context.memoize(obj)
            f_name = f"{obj_id}.npy"
            if f_name not in save_context.zip_file.namelist():
                save_context.zip_file.writestr(f_name, data_buffer.getbuffer())
            res.update(type="numpy", file=f_name)
    except ValueError:
        # Couldn't save the numpy array with either method
        raise UnsupportedTypeException(
            f"numpy arrays of dtype {obj.dtype} are not supported yet, please "
            "open an issue at https://github.com/skops-dev/skops/issues and "
            "report your error"
        )

    return res


class NdArrayNode(Node):
    def __init__(
        self,
        state: dict[str, Any],
        load_context: LoadContext,
        trusted: bool | Sequence[str] = False,
    ) -> None:
        super().__init__(state, load_context, trusted)
        self.type = state["type"]
        self.trusted = self._get_trusted(trusted, [np.ndarray])
        if self.type == "numpy":
            self.children = {
                "content": io.BytesIO(load_context.src.read(state["file"]))
            }
        elif self.type == "json":
            self.children = {
                "content": [  # type: ignore
                    get_tree(o, load_context) for o in state["content"]  # type: ignore
                ],
                "shape": get_tree(state["shape"], load_context),
            }
        else:
            raise ValueError(f"Unknown type {self.type}.")

    def _construct(self):
        # Dealing with a regular numpy array, where dtype != object
        if self.type == "numpy":
            content = np.load(self.children["content"], allow_pickle=False)
            if f"{self.module_name}.{self.class_name}" != "numpy.ndarray":
                content = gettype(self.module_name, self.class_name)(content)
            return content

        if self.type == "json":
            # We explicitly set the dtype to "O" since we only save object
            # arrays in json.
            shape = self.children["shape"].construct()
            tmp = [o.construct() for o in self.children["content"]]

            # TODO: this is a hack to get the correct shape of the array. We
            # should find _a better way_ to do this.
            if len(shape) == 1:
                content = np.ndarray(shape=len(tmp), dtype="O")
                for i, v in enumerate(tmp):
                    content[i] = v
            else:
                content = np.array(tmp, dtype="O")

            return content

        raise ValueError(f"Unknown type for a numpy object: {self.type}.")


def maskedarray_get_state(obj: Any, save_context: SaveContext) -> dict[str, Any]:
    res = {
        "__class__": obj.__class__.__name__,
        "__module__": get_module(type(obj)),
        "__loader__": "MaskedArrayNode",
        "content": {
            "data": get_state(obj.data, save_context),
            "mask": get_state(obj.mask, save_context),
        },
    }
    return res


class MaskedArrayNode(Node):
    def __init__(
        self,
        state: dict[str, Any],
        load_context: LoadContext,
        trusted: bool | Sequence[str] = False,
    ) -> None:
        super().__init__(state, load_context, trusted)
        self.trusted = self._get_trusted(trusted, [np.ma.MaskedArray])
        self.children = {
            "data": get_tree(state["content"]["data"], load_context),
            "mask": get_tree(state["content"]["mask"], load_context),
        }

    def _construct(self):
        data = self.children["data"].construct()
        mask = self.children["mask"].construct()
        return np.ma.MaskedArray(data, mask)


def random_state_get_state(obj: Any, save_context: SaveContext) -> dict[str, Any]:
    content = get_state(obj.get_state(legacy=False), save_context)
    res = {
        "__class__": obj.__class__.__name__,
        "__module__": get_module(type(obj)),
        "__loader__": "RandomStateNode",
        "content": content,
    }
    return res


class RandomStateNode(Node):
    def __init__(
        self,
        state: dict[str, Any],
        load_context: LoadContext,
        trusted: bool | Sequence[str] = False,
    ) -> None:
        super().__init__(state, load_context, trusted)
        self.children = {"content": get_tree(state["content"], load_context)}
        self.trusted = self._get_trusted(trusted, [np.random.RandomState])

    def _construct(self):
        random_state = gettype(self.module_name, self.class_name)()
        random_state.set_state(self.children["content"].construct())
        return random_state


def random_generator_get_state(obj: Any, save_context: SaveContext) -> dict[str, Any]:
    bit_generator_state = get_state(obj.bit_generator.state, save_context)
    res = {
        "__class__": obj.__class__.__name__,
        "__module__": get_module(type(obj)),
        "__loader__": "RandomGeneratorNode",
        "content": {"bit_generator": bit_generator_state},
    }
    return res


class RandomGeneratorNode(Node):
    def __init__(
        self,
        state: dict[str, Any],
        load_context: LoadContext,
        trusted: bool | Sequence[str] = False,
    ) -> None:
        super().__init__(state, load_context, trusted)
        self.children = {
            "bit_generator_state": get_tree(
                state["content"]["bit_generator"], load_context
            )
        }
        self.trusted = self._get_trusted(trusted, [np.random.Generator])

    def _construct(self):
        # first restore the state of the bit generator
        bit_generator_state = self.children["bit_generator_state"].construct()
        bit_generator_cls = gettype(
            "numpy.random", bit_generator_state["bit_generator"]
        )
        bit_generator = bit_generator_cls()
        bit_generator.state = bit_generator_state

        # next create the generator instance
        return gettype(self.module_name, self.class_name)(bit_generator=bit_generator)


def dtype_get_state(obj: Any, save_context: SaveContext) -> dict[str, Any]:
    # we use numpy's internal save mechanism to store the dtype by
    # saving/loading an empty array with that dtype.
    tmp: np.typing.NDArray = np.ndarray(0, dtype=obj)
    res = {
        "__class__": "dtype",
        "__module__": "numpy",
        "__loader__": "DTypeNode",
        "content": get_state(tmp, save_context),
    }
    return res


class DTypeNode(Node):
    def __init__(
        self,
        state: dict[str, Any],
        load_context: LoadContext,
        trusted: bool | Sequence[str] = False,
    ) -> None:
        super().__init__(state, load_context, trusted)
        self.children = {"content": get_tree(state["content"], load_context)}
        # TODO: what should we trust?
        self.trusted = self._get_trusted(trusted, [])

    def _construct(self):
        # we use numpy's internal save mechanism to store the dtype by
        # saving/loading an empty array with that dtype.
        return self.children["content"].construct().dtype


# tuples of type and function that gets the state of that type
GET_STATE_DISPATCH_FUNCTIONS = [
    (np.generic, ndarray_get_state),
    (np.ndarray, ndarray_get_state),
    (np.ma.MaskedArray, maskedarray_get_state),
    (np.ufunc, function_get_state),
    (np.dtype, dtype_get_state),
    (np.random.RandomState, random_state_get_state),
    (np.random.Generator, random_generator_get_state),
]
# tuples of type and function that creates the instance of that type
NODE_TYPE_MAPPING = {
    ("NdArrayNode", PROTOCOL): NdArrayNode,
    ("MaskedArrayNode", PROTOCOL): MaskedArrayNode,
    ("DTypeNode", PROTOCOL): DTypeNode,
    ("RandomStateNode", PROTOCOL): RandomStateNode,
    ("RandomGeneratorNode", PROTOCOL): RandomGeneratorNode,
}



import pytest

from skops.utils.importutils import import_or_raise


@pytest.mark.usefixtures("matplotlib_not_installed")
def test_import_or_raise():
    with pytest.raises(
        ModuleNotFoundError,
        match=(
            "Permutation importance requires matplotlib to be installed. In order"
            " to use permutation importance, you need to install the package in"
            " your current python environment."
        ),
    ):
        import_or_raise("matplotlib", "permutation importance")



from importlib import import_module


def import_or_raise(module, feature_name):
    """Raise error if a given library is not present in the environment.

    Parameters
    ----------
    module: str
        Name of the module.

    feature_name: str
        Name of the feature module is required for.

    Raises
    ------
    ModuleNotFoundError
        Is raised if a given module is not present in the environment
    """
    try:
        module = import_module(module)
    except ImportError as e:
        package = module.split(".")[0]
        raise ModuleNotFoundError(
            f"{feature_name.capitalize()} requires {package} to be installed. In order"
            f" to use {feature_name}, you need to install the package in your current"
            " python environment."
        ) from e
    return module



from __future__ import annotations

import json
import re
import sys
import textwrap
import zipfile
from collections.abc import Mapping
from dataclasses import dataclass, field
from pathlib import Path
from reprlib import Repr
from typing import Any, Iterator, Literal, Sequence, Union

import joblib
from huggingface_hub import ModelCardData
from sklearn.utils import estimator_html_repr
from tabulate import tabulate  # type: ignore

from skops.card._templates import CONTENT_PLACEHOLDER, SKOPS_TEMPLATE, Templates
from skops.io import load
from skops.utils.importutils import import_or_raise

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self


# Repr attributes can be used to control the behavior of repr
aRepr = Repr()
aRepr.maxother = 79
aRepr.maxstring = 79


VALID_TEMPLATES = {item.value for item in Templates}
NEED_SECTION_ERR_MSG = (
    "You are trying to {action} but you're using a custom template, please pass the "
    "'section' argument to determine where to put the content"
)


def wrap_as_details(text: str, folded: bool) -> str:
    if not folded:
        return text
    return f"<details>\n<summary> Click to expand </summary>\n\n{text}\n\n</details>"


def _clean_table(table: str) -> str:
    # replace line breaks "\n" with html tag <br />, however, leave end-of-line
    # line breaks (eol_lb) intact
    eol_lb = "|\n"
    placeholder = "\x1f"  # unit separator control character (ASCII control char 31)
    table = (
        table.replace(eol_lb, placeholder)
        .replace("\n", "<br />")
        .replace(placeholder, eol_lb)
    )
    return table


def metadata_from_config(config_path: Union[str, Path]) -> ModelCardData:
    """Construct a ``ModelCardData`` object from a ``config.json`` file.

    Most information needed for the metadata section of a ``README.md`` file on
    Hugging Face Hub is included in the ``config.json`` file. This utility
    function constructs a :class:`huggingface_hub.ModelCardData` object which
    can then be passed to the :class:`~skops.card.Card` object.

    This method populates the following attributes of the instance:

    - ``library_name``: It needs to be ``"sklearn"`` for scikit-learn
        compatible models.
    - ``tags``: Set to a list, containing ``"sklearn"`` and the task of the
        model. You can then add more tags to this list.
    - ``widget``: It is populated with the example data to be used by the
        widget component of the Hugging Face Hub widget, on the model's
        repository page.

    Parameters
    ----------
    config_path: str, or Path
        Filepath to the ``config.json`` file, or the folder including that
        file.

    Returns
    -------
    card_data: huggingface_hub.ModelCardData
        :class:`huggingface_hub.ModelCardData` object.

    """
    config_path = Path(config_path)
    if not config_path.is_file():
        config_path = config_path / "config.json"

    with open(config_path) as f:
        config = json.load(f)
    card_data = ModelCardData(
        model_format=config.get("sklearn", {}).get("model_format", {})
    )
    card_data.library_name = "sklearn"
    card_data.tags = ["sklearn", "skops"]
    task = config.get("sklearn", {}).get("task", None)
    if task:
        card_data.tags += [task]
    card_data.model_file = config.get("sklearn", {}).get("model", {}).get("file")  # type: ignore
    if config.get("sklearn", {}).get("use_intelex"):
        card_data.tags.append("scikit-learn-intelex")

    example_input = config.get("sklearn", {}).get("example_input", None)
    # Documentation on what the widget expects:
    # https://huggingface.co/docs/hub/models-widgets-examples
    if example_input:
        if "tabular" in task:
            card_data.widget = {"structuredData": example_input}  # type: ignore
        # TODO: add text data example here.

    return card_data


def split_subsection_names(key: str) -> list[str]:
    r"""Split a string containing multiple sections into a list of strings for
    each.

    The separator is ``"/"``. To avoid splitting on ``"/"``, escape it using
    ``"\\/"``.

    Examples
    --------
    >>> split_subsection_names("Section A")
    ['Section A']
    >>> split_subsection_names("Section A/Section B/Section C")
    ['Section A', 'Section B', 'Section C']
    >>> split_subsection_names("A section containing \\/ a slash")
    ['A section containing / a slash']
    >>> split_subsection_names("Spaces are / stripped")
    ['Spaces are', 'stripped']

    Parameters
    ----------
    key : str
        The section name consisting potentially of multiple subsections. It has
        to be ensured beforhand that this is not an empty string.

    Returns
    -------
    parts : list of str
        The individual (sub)sections.

    """
    placeholder = "\x1f"  # unit separator control character (ASCII control char 31)
    key = key.replace("\\/", placeholder)
    parts = (part.strip() for part in key.split("/"))
    return [part.replace(placeholder, "/") for part in parts]


def _getting_started_code(
    file_name: str, model_format: Literal["pickle", "skops"], indent: str = "    "
) -> list[str]:
    # get lines of code required to load the model
    lines = [
        "import json",
        "import pandas as pd",
    ]
    if model_format == "skops":
        lines += ["import skops.io as sio"]
    else:
        lines += ["import joblib"]

    if model_format == "skops":
        lines += [f'model = sio.load("{file_name}")']
    else:  # pickle
        lines += [f'model = joblib.load("{file_name}")']

    lines += [
        'with open("config.json") as f:',
        indent + "config = json.load(f)",
        'model.predict(pd.DataFrame.from_dict(config["sklearn"]["example_input"]))',
    ]
    return lines


@dataclass
class Section:
    """Building block of the model card.

    The model card is represented internally as a dict with keys being strings
    and values being ``Section``s. The key is identical to the section title.

    Additionally, the section may hold content in the form of strings (can be an
    empty string) or a ``Formattable``, which is simply an object with a
    ``format`` method that returns a string.

    The section can contain subsections, which again are dicts of
    string keys and section values (the dict can be empty). Therefore, the model
    card representation forms a tree structure, making use of the fact that dict
    order is preserved.

    The section may also contain a ``visible`` flag, which determines if the
    section will be shown when the card is rendered.

    """

    title: str
    content: str
    subsections: dict[str, Section] = field(default_factory=dict)
    visible: bool = True

    def select(self, key: str) -> Section:
        """Return a subsection or subsubsection of this section

        Parameters
        ----------
        key : str
            The name of the (sub)section to select. When selecting a subsection,
            either use a ``"/"`` in the name to separate the parent and child
            sections, chain multiple ``select`` calls.

        Returns
        -------
        section : Section
            A dataclass containing all information relevant to the selected
            section. Those are the title, the content, subsections (in a dict),
            and additional fields that depend on the type of section.

        Raises
        ------
        KeyError
            If the given section name was not found, a ``KeyError`` is raised.

        """
        section_names = split_subsection_names(key)
        # check that no section name is empty
        if not all(bool(name) for name in section_names):
            msg = f"Section name cannot be empty but got '{key}'"
            raise KeyError(msg)

        section: Section = self
        for section_name in section_names:
            section = section.subsections[section_name]
        return section

    def format(self) -> str:
        return self.content

    def __repr__(self) -> str:
        """Generates the ``repr`` of this section.

        ``repr`` determines how the content of this section is shown in the
        Card's repr.
        """
        return self.content


@dataclass
class PlotSection(Section):
    """Adds a link to a figure to the model card"""

    path: str | Path = ""
    alt_text: str = ""
    folded: bool = False

    def __post_init__(self) -> None:
        if not self.path:
            raise TypeError(f"{self.__class__.__name__} requires a path")

    def format(self) -> str:
        # if no alt text provided, fall back to figure path
        alt_text = self.alt_text or self.path
        text = f"![{alt_text}]({self.path})"
        val = wrap_as_details(text, folded=self.folded)
        if self.content:
            val = f"{self.content}\n\n{val}"
        return val

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.path})"


@dataclass
class TableSection(Section):
    """Adds a table to the model card"""

    table: Mapping[str, Sequence[Any]] = field(default_factory=dict)
    folded: bool = False

    def __post_init__(self) -> None:
        self._check_table()

    def _check_table(self) -> None:
        try:
            import pandas as pd

            self._is_pandas_df = isinstance(self.table, pd.DataFrame)
        except ImportError:
            self._is_pandas_df = False

        if self._is_pandas_df:
            ncols = len(self.table.columns)  # type: ignore
        else:
            ncols = len(self.table)
        if ncols == 0:
            raise ValueError("Trying to add table with no columns")

    def format(self) -> str:
        if self._is_pandas_df:
            headers = self.table.columns  # type: ignore
        else:
            headers = self.table.keys()

        table = _clean_table(
            tabulate(self.table, tablefmt="github", headers=headers, showindex=False)
        )
        val = wrap_as_details(table, folded=self.folded)

        if self.content:
            val = f"{self.content}\n\n{val}"
        return val

    def __repr__(self) -> str:
        if self._is_pandas_df:
            nrows, ncols = self.table.shape  # type: ignore
        else:
            # table cannot be empty, so no checks needed here
            ncols = len(self.table)
            key = next(iter(self.table.keys()))
            nrows = len(self.table[key])
        return f"{self.__class__.__name__}({nrows}x{ncols})"


def _load_model(model: Any, trusted=False) -> Any:
    """Return a model instance.

    Loads the model if provided a file path, if already a model instance return
    it unmodified.

    Parameters
    ----------
    model : pathlib.Path, str, or sklearn estimator
        Path/str or the actual model instance. if a Path or str, loads the model.

    trusted : bool, default=False
        Passed to :func:`skops.io.load` if the model is a file path and it's
        a `skops` file.

    Returns
    -------
    model : object
        Model instance.
    """

    if not isinstance(model, (Path, str)):
        return model

    model_path = Path(model)
    if not model_path.exists():
        raise FileNotFoundError(f"File is not present: {model_path}")

    try:
        if zipfile.is_zipfile(model_path):
            model = load(model_path, trusted=trusted)
        else:
            model = joblib.load(model_path)
    except Exception as ex:
        msg = f'An "{type(ex).__name__}" occurred during model loading.'
        raise RuntimeError(msg) from ex

    return model


class Card:
    """Model card class that will be used to generate model card.

    This class can be used to write information and plots to model card and save
    it. This class by default generates an interactive plot of the model and a
    table of hyperparameters. Some sections are added by default.

    Parameters
    ----------
    model: pathlib.Path, str, or sklearn estimator object
        ``Path``/``str`` of the model or the actual model instance that will be
        documented. If a ``Path`` or ``str`` is provided, model will be loaded.

    model_diagram: bool or "auto" or str, default="auto"
        If using the skops template, setting this to ``True`` or ``"auto"`` will
        add the model diagram, as generated by sckit-learn, to the default
        section, i.e "Model description/Training Procedure/Model Plot". Passing
        a string to ``model_diagram`` will instead use that string as the
        section name for the diagram. Set to ``False`` to not include the model
        diagram.

        If using a non-skops template, passing ``"auto"`` won't add the model
        diagram because there is no pre-defined section to put it. The model
        diagram can, however, always be added later using
        :meth:`Card.add_model_plot`.

    metadata: ModelCardData, optional
        :class:`huggingface_hub.ModelCardData` object. The contents of this
        object are saved as metadata at the beginning of the output file, and
        used by Hugging Face Hub.

        You can use :func:`~skops.card.metadata_from_config` to create an
        instance pre-populated with necessary information based on the contents
        of the ``config.json`` file, which itself is created by
        :func:`skops.hub_utils.init`.

    template: "skops", dict, or None (default="skops")
        Whether to add default sections or not. The template can be a predefined
        template, which at the moment can only be the string ``"skops"``, which
        is a template provided by ``skops`` that is geared towards typical
        sklearn models. If you don't want any prefilled sections, just pass
        ``None``. If you want custom prefilled sections, pass a ``dict``, where
        keys are the sections and values are the contents of the sections. Note
        that when you use no template or a custom template, some methods will
        not work, e.g. :meth:`Card.add_metrics`, since it's not clear where to
        put the metrics when there is no template or a custom template.

    trusted: bool, default=False
        Passed to :func:`skops.io.load` if the model is a file path and it's
        a `skops` file.

    Attributes
    ----------
    model: estimator object
        The scikit-learn compatible model that will be documented.

    metadata: ModelCardData
        Metadata to be stored at the beginning of the saved model card, as
        metadata to be understood by the Hugging Face Hub.

    Examples
    --------
    >>> from sklearn.metrics import (
    ...     ConfusionMatrixDisplay,
    ...     confusion_matrix,
    ...     accuracy_score,
    ...     f1_score
    ... )
    >>> import tempfile
    >>> from pathlib import Path
    >>> from sklearn.datasets import load_iris
    >>> from sklearn.linear_model import LogisticRegression
    >>> from skops.card import Card
    >>> X, y = load_iris(return_X_y=True)
    >>> model = LogisticRegression(solver="liblinear", random_state=0).fit(X, y)
    >>> model_card = Card(model)
    >>> model_card.metadata.license = "mit"
    >>> y_pred = model.predict(X)
    >>> model_card.add_metrics(**{
    ...     "accuracy": accuracy_score(y, y_pred),
    ...     "f1 score": f1_score(y, y_pred, average="micro"),
    ... })
    Card(...)
    >>> cm = confusion_matrix(y, y_pred,labels=model.classes_)
    >>> disp = ConfusionMatrixDisplay(
    ...     confusion_matrix=cm,
    ...     display_labels=model.classes_
    ... )
    >>> disp.plot()
    <sklearn.metrics._plot.confusion_matrix.ConfusionMatrixDisplay object at ...>
    >>> tmp_path = Path(tempfile.mkdtemp(prefix="skops-"))
    >>> disp.figure_.savefig(tmp_path / "confusion_matrix.png")
    ...
    >>> model_card.add_plot(**{
    ...     "Model description/Confusion Matrix": tmp_path / "confusion_matrix.png"
    ... })
    Card(...)
    >>> # add new content to the existing section "Model description"
    >>> model_card.add(**{"Model description": "This is the best model"})
    Card(...)
    >>> # add content to a new section
    >>> model_card.add(**{"A new section": "Please rate my model"})
    Card(...)
    >>> # add new subsection to an existing section by using "/"
    >>> model_card.add(**{"Model description/Model name": "This model is called Bob"})
    Card(
      model=LogisticRegression(random_state=0, solver='liblinear'),
      metadata.license=mit,
      Model description=This is the best model,
      Model description/Training Procedure/Hyperparameters=TableSection(15x2),
      Model description/Training Procedure/...</pre></div></div></div></div></div>,
      Model description/Evaluation Results=TableSection(2x2),
      Model description/Confusion Matrix=Pl...confusion_matrix.png),
      Model description/Model name=This model is called Bob,
      A new section=Please rate my model,
    )
    >>> # save the card to a README.md file
    >>> model_card.save(tmp_path / "README.md")

    """

    def __init__(
        self,
        model,
        model_diagram: bool | Literal["auto"] | str = "auto",
        metadata: ModelCardData | None = None,
        template: Literal["skops"] | dict[str, str] | None = "skops",
        trusted: bool = False,
    ) -> None:
        self.model = model
        self.metadata = metadata or ModelCardData()
        self.template = template
        self.trusted = trusted

        self._data: dict[str, Section] = {}
        self._metrics: dict[str, str | float | int] = {}

        self._populate_template(model_diagram=model_diagram)

    def _populate_template(self, model_diagram: bool | Literal["auto"] | str):
        """If initialized with a template, use it to populate the card.

        Parameters
        ----------
        model_diagram: bool or "auto" or str
            If using the default template, ``"auto"`` and ``True`` will add the
            diagram in its default section. If using a custom template,
            ``"auto"`` will not add the diagram, and passing ``True`` will
            result in an error. For either, passing ``False`` will result in the
            model diagram being omitted, and passing a string (other than
            ``"auto"``) will put the model diagram into a section corresponding
            to that string.

        """
        if isinstance(self.template, str) and (self.template not in VALID_TEMPLATES):
            valid_templates = ", ".join(f"'{val}'" for val in sorted(VALID_TEMPLATES))
            msg = (
                f"Unknown template '{self.template}', "
                f"template must be one of the following values: {valid_templates}"
            )
            raise ValueError(msg)

        # default template
        if self.template == Templates.skops.value:
            self.add(**SKOPS_TEMPLATE)
            # for the skops template, automatically add some default sections
            self.add_hyperparams()
            self.add_get_started_code()

            if (model_diagram is True) or (model_diagram == "auto"):
                self.add_model_plot()
            elif isinstance(model_diagram, str):
                self.add_model_plot(section=model_diagram)
            return

        # non-default template
        if isinstance(self.template, Mapping):
            self.add(**self.template)

        if isinstance(model_diagram, str) and (model_diagram != "auto"):
            self.add_model_plot(section=model_diagram)
        elif model_diagram is True:
            # will trigger an error
            self.add_model_plot()

    def get_model(self) -> Any:
        """Returns sklearn estimator object.

        If the ``model`` is already loaded, return it as is. If the ``model``
        attribute is a ``Path``/``str``, load the model and return it.

        Returns
        -------
        model : BaseEstimator
            The model instance.

        """
        model = _load_model(self.model, self.trusted)
        # Ideally, we would only call the method below if we *know* that the
        # model has changed, but at the moment we have no way of knowing that
        return model

    def add(self, **kwargs: str) -> Self:
        """Add new section(s) to the model card.

        Add one or multiple sections to the model card. The section names are
        taken from the keys and the contents are taken from the values.

        To add to an existing section, use a ``"/"`` in the section name, e.g.:

        ``card.add(**{"Existing section/New section": "content"})``.

        If the parent section does not exist, it will be added automatically.

        To add a section with ``"/"`` in its title (i.e. not inteded as a
        subsection), escape the slash like so, ``"\\/"``, e.g.:

        ``card.add(**{"A section with\\/a slash in the title": "content"})``.

        If a section of the given name already exists, its content will be
        overwritten.

        Parameters
        ----------
        **kwargs : dict
            The keys of the dictionary serve as the section title and the values
            as the section content. It's possible to add to existing sections.

        Returns
        -------
        self : object
            Card object.

        """
        for key, val in kwargs.items():
            self._add_single(key, val)
        return self

    def _select(
        self, subsection_names: Sequence[str], create: bool = True
    ) -> dict[str, Section]:
        """Select a single section from the data.

        Parameters
        ----------
        subsection_names: list of str
            The subsection names, already split into individual subsections.

        create: bool (default=True)
            Whether to create the subsection if it does not already exist or
            not.

        Returns
        -------
        section: dict of Section
            A dict mapping the section key (identical to the title) to the
            actual ``Section``, which is a dataclass that contains the actual
            data of the section.

        Raises
        ------
        KeyError
            If the section does not exist and ``create=False``, raises a
            ``KeyError``.

        """
        section = self._data
        if not subsection_names:
            return section

        for subsection_name in subsection_names:
            section_maybe = section.get(subsection_name)

            # there are already subsections
            if section_maybe is not None:
                section = section_maybe.subsections
                continue

            if create:
                # no subsection, create
                entry = Section(title=subsection_name, content="")
                section[subsection_name] = entry
                section = entry.subsections
            else:
                raise KeyError(f"Section {subsection_name} does not exist")

        return section

    def select(self, key: str) -> Section:
        """Select a section from the model card.

        To select a subsection of an existing section, use a ``"/"`` in the
        section name, e.g.:

        ``card.select("Main section/Subsection")``.

        Alternatively, multiple ``select`` calls can be chained:

        ``card.select("Main section").select("Subsection")``.

        Parameters
        ----------
        key : str
            The name of the (sub)section to select. When selecting a subsection,
            either use a ``"/"`` in the name to separate the parent and child
            sections, chain multiple ``select`` calls.

        Returns
        -------
        self : Section
            A dataclass containing all information relevant to the selected
            section. Those are the title, the content, and subsections (in a
            dict).

        Raises
        ------
        KeyError
            If the given section name was not found, a ``KeyError`` is raised.

        """
        if not key:
            msg = f"Section name cannot be empty but got '{key}'"
            raise KeyError(msg)

        *subsection_names, leaf_node_name = split_subsection_names(key)

        if not leaf_node_name:
            msg = f"Section name cannot be empty but got '{key}'"
            raise KeyError(msg)

        parent_section = self._select(subsection_names, create=False)
        return parent_section[leaf_node_name]

    def delete(self, key: str | Sequence[str]) -> None:
        """Delete a section from the model card.

        To delete a subsection of an existing section, use a ``"/"`` in the
        section name, e.g.:

        ``card.delete("Existing section/New section")``.

        Alternatively, a list of strings can be passed:

        ``card.delete(["Existing section", "New section"])``.

        Parameters
        ----------
        key : str or list of str
            The name of the (sub)section to select. When selecting a subsection,
            either use a ``"/"`` in the name to separate the parent and child
            sections, or pass a list of strings.

        Raises
        ------
        KeyError
            If the given section name was not found, a ``KeyError`` is raised.

        """
        if not key:
            msg = f"Section name cannot be empty but got '{key}'"
            raise KeyError(msg)

        if isinstance(key, str):
            *subsection_names, leaf_node_name = split_subsection_names(key)
        else:
            *subsection_names, leaf_node_name = key

        if not leaf_node_name:
            msg = f"Section name cannot be empty but got '{key}'"
            raise KeyError(msg)

        parent_section = self._select(subsection_names, create=False)
        del parent_section[leaf_node_name]

    def _add_single(self, key: str, val: str | Section) -> Section:
        """Add a single section.

        If the (sub)section does not exist, it is created. Otherwise, the
        existing (sub)section is modified.

        Parameters
        ----------
        key: str
            The name of the (sub)section.

        val: str or Section
            The value to assign to the (sub)section. If this is already a
            section, leave it as it is. If it's a string, create a
            :class:`skops.card._model_card.Section`.

        Returns
        -------
        Section instance
            The section that has been added or modified.

        """
        *subsection_names, leaf_node_name = split_subsection_names(key)
        section = self._select(subsection_names)

        if isinstance(val, str):
            # val is a str, create a Section
            new_section = Section(title=leaf_node_name, content=val)
        else:
            # val is already a section and can be used as is
            new_section = val

        if leaf_node_name in section:
            # entry exists, preserve its subsections
            old_section = section[leaf_node_name]
            if new_section.subsections and (
                new_section.subsections != old_section.subsections
            ):
                msg = (
                    f"Trying to override section '{leaf_node_name}' but found "
                    "conflicting subsections."
                )
                raise ValueError(msg)
            new_section.subsections = old_section.subsections

        section[leaf_node_name] = new_section
        return section[leaf_node_name]

    def add_model_plot(
        self,
        section: str = "Model description/Training Procedure/Model Plot",
        description: str | None = None,
    ) -> Self:
        """Add a model plot

        Use sklearn model visualization to add create a diagram of the model.
        See the `sklearn model visualization docs
        <https://scikit-learn.org/stable/modules/compose.html#visualizing-composite-estimators>`_.

        The model diagram is not added if the card class was instantiated with
        ``model_diagram=False``.

        Parameters
        ----------
        section : str (default="Model description/Training Procedure/Model Plot")
            The section that the model plot should be added to. By default, the
            section is set to fit the skops model card template. If you're using
            a different template, you may have to choose a different section name.

        description : str or None, default=None
            An optional description to be added before the model plot. If you're
            using the default skops template, a standard text is used. Pass a
            string here if you want to use your own text instead. Leave this
            empty to not add any description.

        Returns
        -------
        self : object
            Card object.

        """
        self._add_model_plot(
            self.get_model(), section_name=section, description=description
        )

        return self

    def _add_model_plot(
        self, model: Any, section_name: str, description: str | None
    ) -> None:
        """Add model plot section

        The model should be a loaded sklearn model, not a path.

        """
        model_plot_div = re.sub(r"\n\s+", "", str(estimator_html_repr(model)))
        if model_plot_div.count("sk-top-container") == 1:
            model_plot_div = model_plot_div.replace(
                "sk-top-container", 'sk-top-container" style="overflow: auto;'
            )

        if description:
            content = f"{description}\n\n{model_plot_div}"
        else:
            content = model_plot_div

        description = description or ""
        title = split_subsection_names(section_name)[-1]
        section = Section(title=title, content=content)
        self._add_single(section_name, section)

    def add_hyperparams(
        self,
        section: str = "Model description/Training Procedure/Hyperparameters",
        description: str | None = None,
    ) -> Self:
        """Add the model's hyperparameters as a table

        Parameters
        ----------
        section : str (default="Model description/Training Procedure/Hyperparameters")
            The section that the hyperparameters should be added to. By default,
            the section is set to fit the skops model card template. If you're
            using a different template, you may have to choose a different section
            name.

        description : str or None, default=None
            An optional description to be added before the hyperparamters. If
            you're using the default skops template, a standard text is used.
            Pass a string here if you want to use your own text instead. Leave
            this empty to not add any description.

        Returns
        -------
        self : object
            Card object.

        """
        self._add_hyperparams(
            self.get_model(), section_name=section, description=description
        )
        return self

    def _add_hyperparams(
        self, model: Any, section_name: str, description: str | None
    ) -> None:
        """Add hyperparameter section.

        The model should be a loaded sklearn model, not a path.

        """
        params = model.get_params(deep=True)
        table = {"Hyperparameter": list(params.keys()), "Value": list(params.values())}

        description = description or ""
        title = split_subsection_names(section_name)[-1]
        section = TableSection(
            title=title, content=description, table=table, folded=True
        )
        self._add_single(section_name, section)

    def add_get_started_code(
        self,
        section: str = "How to Get Started with the Model",
        description: str | None = None,
        file_name: str | None = None,
        model_format: Literal["pickle", "skops"] | None = None,
    ) -> Self:
        """Add getting started code

        This code can be copied by users to load the model and make predictions
        with it.

        Parameters
        ----------
        section : str (default="How to Get Started with the Model")
            The section that the code for loading the model should be added to.
            By default, the section is set to fit the skops model card template.
            If you're using a different template, you may have to choose a
            different section name.

        description : str or None, default=None
            An optional description to be added before the code. If you're using
            the default skops template, a standard text is used. Pass a string
            here if you want to use your own text instead. Leave this empty to
            not add any description.

        file_name : str or None, default=None
            The file name of the model. If no file name is indicated, there will
            be an attempt to read the file name from the card's metadata. If
            that fails, an error is raised and you have to pass this argument
            explicitly.

        model_format : "skops", "pickle", or None, default=None
            The model format used to store the model.If format is indicated,
            there will be an attempt to read the model format from the card's
            metadata. If that fails, an error is raised and you have to pass
            this argument explicitly.

        Returns
        -------
        self : object
            Card object.

        """
        if file_name is None:
            file_name = self.metadata.to_dict().get("model_file")

        if model_format is None:
            model_format = (
                self.metadata.to_dict().get("sklearn", {}).get("model_format")
            )

        if model_format and (model_format not in ("pickle", "skops")):
            msg = (
                f"Invalid model format '{model_format}', should be one of "
                "'pickle' or 'skops'"
            )
            raise ValueError(msg)

        if (not file_name) or (not model_format):
            return self

        self._add_get_started_code(
            section,
            file_name=file_name,
            model_format=model_format,
            description=description,
        )

        return self

    def _add_get_started_code(
        self,
        section_name: str,
        file_name: str,
        model_format: Literal["pickle", "skops"],
        description: str | None,
        indent: str = "    ",
    ) -> None:
        """Add getting started code to the corresponding section"""
        lines = _getting_started_code(
            file_name, model_format=model_format, indent=indent
        )
        lines = ["```python"] + lines + ["```"]
        code = "\n".join(lines)

        if description:
            content = f"{description}\n\n{code}"
        else:
            content = code

        title = split_subsection_names(section_name)[-1]
        section = Section(title=title, content=content)
        self._add_single(section_name, section)

    def add_plot(
        self,
        *,
        description: str | None = None,
        alt_text: str | None = None,
        folded=False,
        **kwargs: str | Path,
    ) -> Self:
        """Add plots to the model card.

        The plot should be saved on the file system and the path passed as
        value.

        Parameters
        ----------
        description: str or None (default=None)
            If a string is passed as description, it is shown before the figure.
            If multiple figures are added with one call, they all get the same
            description. To add multiple figures with different descriptions,
            call this method multiple times.

        alt_text: : str or None (default=None)
            If a string is passed as ``alt_text``, it is used as the alternative
            text for the figure (i.e. what is shown if the figure cannot be
            rendered). If this argument is ``None``, the alt_text will just be
            the same as the section title. If multiple figures are added with
            one call, they all get the same alt text. To add multiple figures
            with different alt texts, call this method multiple times.

        folded: bool (default=False)
            If set to ``True``, the plot will be enclosed in a ``details`` tag.
            That means the content is folded by default and users have to click
            to show the content. This option is useful if the added plot is
            large.

        **kwargs : dict
            The arguments should be of the form ``name=plot_path``, where
            ``name`` is the name of the plot and section, and ``plot_path`` is
            the path to the plot on the file system (either a str or
            ``pathlib.Path``), relative to the root of the project. The plots
            should have already been saved under the project's folder.

        Returns
        -------
        self : object
            Card object.

        """
        description = description or ""
        for section_name, plot_path in kwargs.items():
            title = split_subsection_names(section_name)[-1]
            alt_text = alt_text or title
            section = PlotSection(
                title=title,
                content=description,
                alt_text=alt_text,
                path=plot_path,
                folded=folded,
            )
            self._add_single(section_name, section)
        return self

    def add_table(
        self,
        *,
        description: str | None = None,
        folded: bool = False,
        **kwargs: dict["str", list[Any]],
    ) -> Self:
        """Add a table to the model card.

        Add a table to the model card. This can be especially useful when you
        using cross validation with sklearn. E.g. you can directly pass the
        result from calling :func:`sklearn.model_selection.cross_validate` or
        the ``cv_results_`` attribute from any of the hyperparameter searches,
        such as :class:`sklearn.model_selection.GridSearchCV`.

        Morevoer, you can pass any pandas :class:`pandas.DataFrame` to this
        method and it will be rendered in the model card. You may consider
        selecting only a part of the table if it's too big:

        .. code:: python

            search = GridSearchCV(...)
            search.fit(X, y)
            df = pd.DataFrame(search.cv_results_)
            # show only top 10 highest scores
            df = df.sort_values(["mean_test_score"], ascending=False).head(10)
            model_card = skops.card.Card(...)
            model_card.add_table(**{"Hyperparameter search results top 10": df})

        Parameters
        ----------
        description: str or None (default=None)
            If a string is passed as description, it is shown before the table.
            If multiple tables are added with one call, they all get the same
            description. To add multiple tables with different descriptions,
            call this method multiple times.

        folded: bool (default=False)
            If set to ``True``, the table will be enclosed in a ``details`` tag.
            That means the content is folded by default and users have to click
            to show the content. This option is useful if the added table is
            large.

        **kwargs : dict
            The keys should be strings, which will be used as the section
            headers, and the values should be tables. Tables can be either dicts
            with the key being strings that represent the column name, and the
            values being lists that represent the entries for each row.
            Alternatively, the table can be a :class:`pandas.DataFrame`. The
            table must not be empty.

        Returns
        -------
        self : object
            Card object.

        """
        description = description or ""
        for key, val in kwargs.items():
            section = TableSection(
                title=key, content=description, table=val, folded=folded
            )
            self._add_single(key, section)
        return self

    def add_metrics(
        self,
        section: str = "Model description/Evaluation Results",
        description: str | None = None,
        **kwargs: str | int | float,
    ) -> Self:
        """Add metric values to the model card.

        All metrics will be collected in, and then formatted to, a table.

        Parameters
        ----------
        section : str (default="Model description/Evaluation Results")
            The section that metrics should be added to. By default, the section
            is set to fit the skops model card template. If you're using a
            different template, you may have to choose a different section name.

        description : str or None, default=None
            An optional description to be added before the metrics. If you're
            using the default skops template, a standard text is used. Pass a
            string here if you want to use your own text instead. Leave this
            empty to not add any description.

        **kwargs : dict
            A dictionary of the form ``{metric name: metric value}``.

        Returns
        -------
        self : object
            Card object.

        """
        self._metrics.update(kwargs)
        self._add_metrics(section, description=description, metrics=self._metrics)
        return self

    def add_permutation_importances(
        self,
        permutation_importances,
        columns: Sequence[str],
        plot_file: str | Path = "permutation_importances.png",
        plot_name: str = "Permutation Importances",
        overwrite: bool = False,
        description: str | None = None,
    ) -> Self:
        """Plots permutation importance and saves it to model card.

        Parameters
        ----------
        permutation_importances : sklearn.utils.Bunch
            Output of :func:`sklearn.inspection.permutation_importance`.

        columns : str, list or pandas.Index
            Column names of the data used to generate importances.

        plot_file : str or pathlib.Path
            Filename for the plot.

        plot_name : str
            Name of the plot.

        overwrite : bool (default=False)
            Whether to overwrite the permutation importance plot file, if a plot by that
            name already exists.

        description : str | None (default=None)
            An optional description to be added before the plot.

        Returns
        -------
        self : object
            Card object.
        """
        plt = import_or_raise("matplotlib.pyplot", "permutation importance")

        if Path(plot_file).exists() and overwrite is False:
            raise ValueError(
                f"{str(plot_file)} already exists. Set `overwrite` to `True` or pass a"
                " different filename for the plot."
            )
        sorted_importances_idx = permutation_importances.importances_mean.argsort()
        _, ax = plt.subplots()
        ax.boxplot(
            x=permutation_importances.importances[sorted_importances_idx].T,
            labels=columns[sorted_importances_idx],
            vert=False,
        )
        ax.set_title(plot_name)
        ax.set_xlabel("Decrease in Score")
        plt.savefig(plot_file)
        self.add_plot(description=description, alt_text=None, **{plot_name: plot_file})

        return self

    def add_fairlearn_metric_frame(
        self,
        metric_frame,
        table_name: str = "Fairlearn MetricFrame Table",
        transpose: bool = True,
        description: str | None = None,
    ) -> Self:
        """
        Add a :class:`fairlearn.metrics.MetricFrame` table to the model card. The table contains
        the difference, group_ma, group_min, and ratio for each metric.

        Parameters
        ----------
        metric_frame: MetricFrame
            The Fairlearn MetricFrame to add to the model card.

        table_name: str
            The desired name of the table section in the model card.

        transpose: bool, default=True
            Whether to transpose the table or not.


        description : str | None (default=None)
            An optional description to be added before the table.

        Returns
        -------
        self: Card
            The model card with the metric frame added.

        Notes
        --------
        You can check `fairlearn's documentation
        <https://fairlearn.org/v0.8/user_guide/assessment/index.html>`__ on how to
        work with `MetricFrame`s.

        """
        frame_dict = {
            "difference": metric_frame.difference(),
            "group_max": metric_frame.group_max(),
            "group_min": metric_frame.group_min(),
            "ratio": metric_frame.ratio(),
        }

        if transpose is True:
            pd = import_or_raise("pandas", "Pandas is used to pivot the table.")

            frame_dict = pd.DataFrame(frame_dict).T

        return self.add_table(
            folded=True, description=description, **{table_name: frame_dict}
        )

    def _add_metrics(
        self,
        section_name: str,
        description: str | None,
        metrics: dict[str, str | float | int],
    ) -> None:
        """Add metrics to the Evaluation Results section."""
        if self._metrics:
            # transpose from row oriented to column oriented
            data_transposed = zip(*self._metrics.items())
            table = {
                key: list(val) for key, val in zip(["Metric", "Value"], data_transposed)
            }
        else:
            # create empty table
            table = {"Metric": [], "Value": []}

        description = description or ""
        title = split_subsection_names(section_name)[-1]
        section = TableSection(title=title, content=description, table=table)
        self._add_single(section_name, section)

    def _generate_metadata(self, metadata: ModelCardData) -> Iterator[str]:
        """Yield metadata in yaml format"""
        for key, val in metadata.to_dict().items() if metadata else {}:
            yield aRepr.repr(f"metadata.{key}={val},").strip('"').strip("'")

    def _generate_content(
        self, data: dict[str, Section], depth: int = 1
    ) -> Iterator[str]:
        """Yield title and (formatted) contents.

        Recursively go through the data and consecutively yield the title with
        the appropriate number of "#"s (markdown format), then the associated
        content.

        """
        for section in data.values():
            if not section.visible:
                continue

            title = f"{depth * '#'} {section.title}"
            yield title

            yield section.format()

            if section.subsections:
                yield from self._generate_content(section.subsections, depth=depth + 1)

    def _iterate_content(
        self, data: dict[str, Section], parent_section: str = ""
    ) -> Iterator[tuple[str, Section]]:
        """Yield tuples of title and (non-formatted) content."""
        for val in data.values():
            if parent_section:
                title = "/".join((parent_section, val.title))
            else:
                title = val.title

            yield title, val

            if val.subsections:
                yield from self._iterate_content(val.subsections, parent_section=title)

    @staticmethod
    def _format_repr(text: str) -> str:
        # Remove new lines, multiple spaces, quotation marks, and cap line length
        text = text.replace("\n", " ")
        text = re.sub(r"\s+", r" ", text)
        return aRepr.repr(text).strip('"').strip("'")

    def __str__(self) -> str:
        return self.__repr__()

    def __repr__(self) -> str:
        # repr for the model
        model = getattr(self, "model", None)
        if model:
            model_repr = self._format_repr(f"model={repr(self.get_model())},")
        else:
            model_repr = None

        # repr for metadata
        metadata_reprs = []
        for key, val in self.metadata.to_dict().items() if self.metadata else {}:
            if key == "widget":
                metadata_reprs.append("metadata.widget={...},")
                continue

            metadata_reprs.append(self._format_repr(f"metadata.{key}={val},"))
        metadata_repr = "\n".join(metadata_reprs)

        # repr for contents
        content_reprs = []
        for title, section in self._iterate_content(self._data):
            content = section.format()
            if not content:
                continue
            if content.rstrip("`").rstrip().endswith(CONTENT_PLACEHOLDER):
                # if content is just some default text, no need to show it
                continue
            content_reprs.append(self._format_repr(f"{title}={section},"))
        content_repr = "\n".join(content_reprs)

        # combine all parts
        complete_repr = "Card(\n"
        if model_repr:
            complete_repr += textwrap.indent(model_repr, "  ") + "\n"
        if metadata_reprs:
            complete_repr += textwrap.indent(metadata_repr, "  ") + "\n"
        if content_reprs:
            complete_repr += textwrap.indent(content_repr, "  ") + "\n"
        complete_repr += ")"
        return complete_repr

    def _generate_card(self) -> Iterator[str]:
        """Yield sections of the model card, including the metadata."""
        if self.metadata.to_dict():
            yield f"---\n{self.metadata.to_yaml()}\n---"

        for line in self._generate_content(self._data):
            if line:
                yield "\n" + line

        # add an empty line add the end
        yield ""

    def save(self, path: str | Path) -> None:
        """Save the model card.

        This method renders the model card in markdown format and then saves it
        as the specified file.

        Parameters
        ----------
        path: str, or Path
            Filepath to save your card.

        Notes
        -----
        The keys in model card metadata can be seen `here
        <https://huggingface.co/docs/hub/models-cards#model-card-metadata>`__.
        """
        with open(path, "w", encoding="utf-8") as f:
            f.write("\n".join(self._generate_card()))

    def render(self) -> str:
        """Render the final model card as a string.

        Returns
        -------
        result : str
            The rendered model card with all placeholders filled and all extra
            sections inserted.
        """
        return "\n".join(self._generate_card())

    def _iterate_key_section_content(
        self,
        data: dict[str, Section],
        level: int = 0,
    ):
        """Iterate through the key sections and yield the title and level.

        Parameters
        ----------
        data : dict[str, Section]
            The card data to iterate through. This is usually the sections and subsections.

        level : int, optional
            The level of the section, by default 0. This keeps track of subsections.

        Returns
        -------
        table_of_contents : str
        """
        for key, val in data.items():
            if not getattr(val, "visible", True):
                continue

            title = val.title
            yield title, level

            if val.subsections:
                yield from self._iterate_key_section_content(
                    val.subsections,
                    level=level + 1,
                )

    def get_toc(self) -> str:
        """Get the table of contents for the model card.

        Returns
        -------
        toc : str
            The table of contents for the model card formatted as a markdown string.
            Example:
                - Model description
                    - Intended uses & limitations
                    - Training Procedure
                        - Hyperparameters
                        - Model Plot
                    - Evaluation Results
                - How to Get Started with the Model
                - Model Card Authors
                - Model Card Contact
        """
        sections = []
        for title, level in self._iterate_key_section_content(self._data):
            sections.append(f"{'  ' * level}- {title}")

        return "\n".join(sections)



from __future__ import annotations

import io
from contextlib import contextmanager
from typing import Any, Generator, Literal, Sequence, Type, Union

from ._protocol import PROTOCOL
from ._trusted_types import PRIMITIVE_TYPE_NAMES
from ._utils import LoadContext, get_module, get_type_paths
from .exceptions import UntrustedTypesFoundException

NODE_TYPE_MAPPING: dict[tuple[str, int], Node] = {}


def check_type(
    module_name: str, type_name: str, trusted: Literal[True] | Sequence[str]
) -> bool:
    """Check if a type is safe to load.

    A type is safe to load only if it's present in the trusted list.

    Parameters
    ----------
    module_name : str
        The module name of the type.

    type_name : str
        The class name of the type.

    trusted : True, or list of str
        If ``True``, the tree is considered safe. Otherwise trusted has to be
        a list of trusted types.

    Returns
    -------
    is_safe : bool
        True if the type is safe, False otherwise.
    """
    if trusted is True:
        return True
    return module_name + "." + type_name in trusted


def audit_tree(tree: Node, trusted: bool | Sequence[str]) -> None:
    """Audit a tree of nodes.

    A tree is safe if it only contains trusted types. Audit is skipped if
    trusted is ``True``.

    Parameters
    ----------
    tree : skops.io._dispatch.Node
        The tree to audit.

    trusted : True, or list of str
        If ``True``, the tree is considered safe. Otherwise trusted has to be
        a list of trusted types names.

        An entry in the list is typically of the form
        ``skops.io._utils.get_module(obj) + "." + obj.__class__.__name__``.

    Raises
    ------
    UntrustedTypesFoundException
        If the tree contains an untrusted type.
    """
    if trusted is True:
        return

    unsafe = tree.get_unsafe_set()
    if isinstance(trusted, (list, set)):
        unsafe -= set(trusted)
    if unsafe:
        raise UntrustedTypesFoundException(unsafe)


class UNINITIALIZED:
    """Sentinel value to indicate that a value has not been initialized yet."""


# Node: types for Generator mean: YieldType, SendType, ReturnType
@contextmanager
def temp_setattr(obj: Any, **kwargs: Any) -> Generator[None, None, None]:
    """Context manager to temporarily set attributes on an object."""
    existing_attrs = {k for k in kwargs.keys() if hasattr(obj, k)}
    previous_values = {k: getattr(obj, k, None) for k in kwargs}
    for k, v in kwargs.items():
        setattr(obj, k, v)
    try:
        yield
    finally:
        for k, v in previous_values.items():
            if k in existing_attrs:
                setattr(obj, k, v)
            else:
                delattr(obj, k)


class Node:
    """A node in the tree of objects.

    This class is a parent class for all nodes in the tree of objects. Each
    type of object (e.g. dict, list, etc.) has its own subclass of Node.

    Each child class has to implement two methods: ``__init__`` and
    ``_construct``.

    ``__init__`` takes care of traversing the state tree and to create the
    corresponding ``Node`` objects. It has access to the ``load_context`` which
    in turn has access to the source zip file. The child class's ``__init__``
    must load attributes into the ``children`` attribute, which is a
    dictionary of ``{child_name: unloaded_value/Node/list/etc}``. The
    ``get_unsafe_set`` should be able to parse and validate the values set
    under the ``children`` attribute. Note that primitives are persisted as a
    ``JsonNode``.

    ``_construct`` takes care of constructing the object. It is only called
    once and the result is cached in ``construct`` which is implemented in this
    class. All required data to construct an instance should be loaded during
    ``__init__``.

    The separation of ``__init__`` and ``_construct`` is necessary because
    audit methods are called after ``__init__`` and before ``construct``.
    Therefore ``__init__`` should avoid creating any instances or importing
    any modules, to avoid running potentially untrusted code.

    Parameters
    ----------
    state : dict
        A dict representing the state of the dumped object.

    load_context : LoadContext
        The context of the loading process.

    trusted : bool or list of str, default=False
        If ``True``, the object will be loaded without any security checks. If
        ``False``, the object will be loaded only if there are only trusted
        objects in the dumped file. If a list of strings, the object will be
        loaded only if all of its required types are listed in ``trusted``
        or are trusted by default.

    memoize : bool, default=True
        If ``True``, the object will be memoized in the load context, if it has
        the ``__id__`` set. This is used to avoid loading the same object
        multiple times.
    """

    def __init__(
        self,
        state: dict[str, Any],
        load_context: LoadContext,
        trusted: bool | Sequence[str] = False,
        memoize: bool = True,
    ) -> None:
        self.class_name, self.module_name = state["__class__"], state["__module__"]
        self._is_safe = None
        self._constructed = UNINITIALIZED
        saved_id = state.get("__id__")
        if saved_id and memoize:
            # hold reference to obj in case same instance encountered again in
            # save state
            load_context.memoize(self, saved_id)

        # subclasses should always:
        # 1. call super().__init__()
        # 2. set self.trusted = self._get_trusted(trusted, ...) where ... is a
        #    list of appropriate trusted types
        # 3. set self.children, where children are states of child nodes; do not
        #    construct the children objects yet
        self.trusted = self._get_trusted(trusted, [])
        self.children: dict[str, Any] = {}

    def construct(self):
        """Construct the object.

        We only construct the object once, and then cache the result.
        """
        if self._constructed is not UNINITIALIZED:
            return self._constructed
        self._constructed = self._construct()
        return self._constructed

    def _construct(self):
        raise NotImplementedError(
            f"{self.__class__.__name__} should implement a '_construct' method"
        )

    @staticmethod
    def _get_trusted(
        trusted: bool | Sequence[Union[str, Type]], default: Sequence[Union[str, Type]]
    ) -> Literal[True] | list[str]:
        """Return a trusted list, or True.

        If ``trusted`` is ``False``, we return the ``default``, otherwise the
        ``trusted`` value is used.

        This is a convenience method called by child classes.
        """
        if trusted is True:
            # if trusted is True, we trust the node
            return True

        if trusted is False:
            # if trusted is False, we only trust the defaults
            return get_type_paths(default)

        # otherwise, we trust the given list
        return get_type_paths(trusted)

    def is_self_safe(self) -> bool:
        """True only if the node's type is considered safe.

        This property only checks the type of the node, not its children.
        """
        return check_type(self.module_name, self.class_name, self.trusted)

    def is_safe(self) -> bool:
        """True only if the node and all its children are safe."""
        # if trusted is set to True, we don't do any safety checks.
        if self.trusted is True:
            return True

        return len(self.get_unsafe_set()) == 0

    def get_unsafe_set(self) -> set[str]:
        """Get the set of unsafe types.

        This method returns all types which are not trusted, including this
        node and all its children.

        Returns
        -------
        unsafe_set : set
            A set of unsafe types.
        """
        if hasattr(self, "_computing_unsafe_set"):
            # this means we're already computing this node's unsafe set, so we
            # return an empty set and let the computation of the parent node
            # continue. This is to avoid infinite recursion.
            return set()

        with temp_setattr(self, _computing_unsafe_set=True):
            res = set()
            if not self.is_self_safe():
                res.add(self.module_name + "." + self.class_name)

            for child in self.children.values():
                if child is None:
                    continue

                # Get the safety set based on the type of the child. In most cases
                # other than ListNode and DictNode, children are all of type Node.
                if isinstance(child, list):
                    # iterate through the list
                    for value in child:
                        res.update(value.get_unsafe_set())
                elif isinstance(child, dict):
                    # iterate through the values of the dict only
                    # TODO: should we check the types of the keys?
                    for value in child.values():
                        res.update(value.get_unsafe_set())
                elif isinstance(child, Node):
                    # delegate to the child Node
                    res.update(child.get_unsafe_set())
                elif type(child) is type:
                    # the if condition bellow is not merged with the previous
                    # one because if the above condition is True, the following
                    # conditions about BytesIO, etc should be ignored.
                    if not check_type(get_module(child), child.__name__, self.trusted):
                        # if the child is a type, we check its safety
                        res.add(get_module(child) + "." + child.__name__)
                elif isinstance(child, io.BytesIO):
                    # We trust BytesIO objects, which are read by other
                    # libraries such as numpy, scipy.
                    continue
                elif check_type(
                    get_module(child), child.__class__.__name__, PRIMITIVE_TYPE_NAMES
                ):
                    # if the child is a primitive type, we don't need to check its
                    # safety.
                    continue
                else:
                    raise ValueError(
                        f"Cannot determine the safety of type {type(child)}. Please"
                        " open an issue at https://github.com/skops-dev/skops/issues"
                        " for us to fix the issue."
                    )

        return res

    def format(self) -> str:
        """Representation of the node's content."""
        return f"{self.module_name}.{self.class_name}"


class CachedNode(Node):
    def __init__(
        self,
        state: dict[str, Any],
        load_context: LoadContext,
        trusted: bool = False,
    ):
        # we pass memoize as False because we don't want to memoize the cached
        # node.
        super().__init__(state, load_context, trusted, memoize=False)
        self.trusted = True
        # TODO: deal with case that __id__ is unknown or prevent it from
        # happening
        self.cached = load_context.get_object(state.get("__id__"))  # type: ignore
        self.children = {}  # type: ignore

    def _construct(self):
        # TODO: FIXME This causes a recursion error when loading a cached
        # object if we call the cached object's `construct``. Some refactoring
        # is needed to fix this.
        return self.cached.construct()


NODE_TYPE_MAPPING[("CachedNode", PROTOCOL)] = CachedNode  # type: ignore


def get_tree(state: dict[str, Any], load_context: LoadContext) -> Node:
    """Get the tree of nodes.

    This function returns the root node of the tree of nodes. The tree is
    constructed recursively by traversing the state tree. No instances are
    created during this process. One would need to call ``construct`` on the
    root node to create the instances.

    This function also handles memoization of the nodes. If a node has already
    been created, it is returned instead of creating a new one.

    Parameters
    ----------
    state : dict
        The state of the dumped object.

    load_context : LoadContext
        The context of the loading process.

    Returns
    -------
    loaded_tree : Node
        The tree containing all its (non-instantiated) child nodes.
    """
    saved_id = state.get("__id__")
    if saved_id in load_context.memo:
        # This means the node is already loaded, so we return it. Note that the
        # node is not constructed at this point. It will be constructed when
        # the parent node's ``construct`` method is called, and for this node
        # it'll be called more than once. But that's not an issue since the
        # node's ``construct`` method caches the instance.
        return load_context.get_object(saved_id)

    loader: str = state["__loader__"]
    protocol = load_context.protocol
    key = (loader, protocol)

    if key in NODE_TYPE_MAPPING:
        node_cls = NODE_TYPE_MAPPING[key]
    else:
        # What probably happened here is that we released a new protocol. If
        # there is no specific key for the old protocol, it means it is safe to
        # use the current protocol instead, because this node was not changed.
        key_new = (loader, PROTOCOL)
        try:
            node_cls = NODE_TYPE_MAPPING[key_new]
        except KeyError:
            # If we still cannot find the loader for this key, something went
            # wrong.
            type_name = f"{state['__module__']}.{state['__class__']}"
            raise TypeError(
                f" Can't find loader {state['__loader__']} for type {type_name} and "
                f"protocol {protocol}."
            )

    loaded_tree = node_cls(state, load_context, trusted=False)  # type: ignore
    return loaded_tree



from __future__ import annotations

from typing import Any, Callable, Sequence, Type

from sklearn.cluster import Birch

from ._protocol import PROTOCOL

try:
    # TODO: remove once support for sklearn<1.2 is dropped. See #187
    from sklearn.covariance._graph_lasso import _DictWithDeprecatedKeys
except ImportError:
    _DictWithDeprecatedKeys = None
from sklearn.linear_model._sgd_fast import (
    EpsilonInsensitive,
    Hinge,
    Huber,
    Log,
    LossFunction,
    ModifiedHuber,
    SquaredEpsilonInsensitive,
    SquaredHinge,
    SquaredLoss,
)
from sklearn.tree._tree import Tree

from ._audit import Node, get_tree
from ._general import unsupported_get_state
from ._utils import LoadContext, SaveContext, get_module, get_state, gettype
from .exceptions import UnsupportedTypeException

ALLOWED_SGD_LOSSES = {
    ModifiedHuber,
    Hinge,
    SquaredHinge,
    Log,
    SquaredLoss,
    Huber,
    EpsilonInsensitive,
    SquaredEpsilonInsensitive,
}

UNSUPPORTED_TYPES = {Birch}


def reduce_get_state(obj: Any, save_context: SaveContext) -> dict[str, Any]:
    # This method is for objects for which we have to use the __reduce__
    # method to get the state.
    res = {
        "__class__": obj.__class__.__name__,
        "__module__": get_module(type(obj)),
    }

    # We get the output of __reduce__ and use it to reconstruct the object.
    # For security reasons, we don't save the constructor object returned by
    # __reduce__, and instead use the pre-defined constructor for the object
    # that we know. This avoids having a function such as `eval()` as the
    # "constructor", abused by attackers.
    #
    # We can/should also look into removing __reduce__ from scikit-learn,
    # and that is not impossible. Most objects which use this don't really
    # need it.
    #
    # More info on __reduce__:
    # https://docs.python.org/3/library/pickle.html#object.__reduce__
    #
    # As a good example, this makes Tree object to be serializable.
    reduce = obj.__reduce__()
    res["__reduce__"] = {}
    res["__reduce__"]["args"] = get_state(reduce[1], save_context)

    if len(reduce) == 3:
        # reduce includes what's needed for __getstate__ and we don't need to
        # call __getstate__ directly.
        attrs = reduce[2]
    elif hasattr(obj, "__getstate__"):
        # since python311 __getstate__ is defined for `object` and might return
        # None
        attrs = obj.__getstate__() or {}
    elif hasattr(obj, "__dict__"):
        attrs = obj.__dict__
    else:
        attrs = {}

    if not isinstance(attrs, (dict, tuple)):
        raise UnsupportedTypeException(
            f"Objects of type {res['__class__']} not supported yet"
        )

    res["content"] = get_state(attrs, save_context)
    return res


class ReduceNode(Node):
    def __init__(
        self,
        state: dict[str, Any],
        load_context: LoadContext,
        constructor: Type[Any] | Callable[..., Any],
        trusted: bool | Sequence[str] = False,
    ) -> None:
        super().__init__(state, load_context, trusted)
        reduce = state["__reduce__"]
        self.children = {
            "attrs": get_tree(state["content"], load_context),
            "args": get_tree(reduce["args"], load_context),
            "constructor": constructor,
        }

    def _construct(self):
        args = self.children["args"].construct()
        constructor = self.children["constructor"]
        instance = constructor(*args)
        attrs = self.children["attrs"].construct()
        if not attrs:
            # nothing more to do
            return instance

        if isinstance(args, tuple) and not hasattr(instance, "__setstate__"):
            raise UnsupportedTypeException(
                f"Objects of type {constructor} are not supported yet"
            )

        if hasattr(instance, "__setstate__"):
            instance.__setstate__(attrs)
        elif isinstance(attrs, dict):
            instance.__dict__.update(attrs)
        else:
            # we (probably) got tuple attrs but cannot setstate with them
            raise UnsupportedTypeException(
                f"Objects of type {constructor} are not supported yet"
            )

        return instance


def tree_get_state(obj: Any, save_context: SaveContext) -> dict[str, Any]:
    state = reduce_get_state(obj, save_context)
    state["__loader__"] = "TreeNode"
    return state


class TreeNode(ReduceNode):
    def __init__(
        self,
        state: dict[str, Any],
        load_context: LoadContext,
        trusted: bool | Sequence[str] = False,
    ) -> None:
        super().__init__(state, load_context, constructor=Tree, trusted=trusted)
        self.trusted = self._get_trusted(trusted, [get_module(Tree) + ".Tree"])


def sgd_loss_get_state(obj: Any, save_context: SaveContext) -> dict[str, Any]:
    state = reduce_get_state(obj, save_context)
    state["__loader__"] = "SGDNode"
    return state


class SGDNode(ReduceNode):
    def __init__(
        self,
        state: dict[str, Any],
        load_context: LoadContext,
        trusted: bool | Sequence[str] = False,
    ) -> None:
        # TODO: make sure trusted here makes sense and used.
        super().__init__(
            state,
            load_context,
            constructor=gettype(state["__module__"], state["__class__"]),
            trusted=False,
        )
        self.trusted = self._get_trusted(
            trusted, [get_module(x) + "." + x.__name__ for x in ALLOWED_SGD_LOSSES]
        )


# TODO: remove once support for sklearn<1.2 is dropped.
def _DictWithDeprecatedKeys_get_state(
    obj: Any, save_context: SaveContext
) -> dict[str, Any]:
    res = {
        "__class__": obj.__class__.__name__,
        "__module__": get_module(type(obj)),
        "__loader__": "_DictWithDeprecatedKeysNode",
    }
    content = {}
    # explicitly pass a dict object instead of _DictWithDeprecatedKeys and
    # later construct a _DictWithDeprecatedKeys object.
    content["main"] = get_state(dict(obj), save_context)
    content["_deprecated_key_to_new_key"] = get_state(
        obj._deprecated_key_to_new_key, save_context
    )
    res["content"] = content
    return res


# TODO: remove once support for sklearn<1.2 is dropped.
class _DictWithDeprecatedKeysNode(Node):
    # _DictWithDeprecatedKeys is just a wrapper for dict
    def __init__(
        self,
        state: dict[str, Any],
        load_context: LoadContext,
        trusted: bool | Sequence[str] = False,
    ) -> None:
        super().__init__(state, load_context, trusted)
        self.trusted = [
            get_module(_DictWithDeprecatedKeysNode) + "._DictWithDeprecatedKeys"
        ]
        self.children = {
            "main": get_tree(state["content"]["main"], load_context),
            "_deprecated_key_to_new_key": get_tree(
                state["content"]["_deprecated_key_to_new_key"], load_context
            ),
        }

    def _construct(self):
        instance = _DictWithDeprecatedKeys(**self.children["main"].construct())
        instance._deprecated_key_to_new_key = self.children[
            "_deprecated_key_to_new_key"
        ].construct()
        return instance


# tuples of type and function that gets the state of that type
GET_STATE_DISPATCH_FUNCTIONS = [
    (LossFunction, sgd_loss_get_state),
    (Tree, tree_get_state),
]
for type_ in UNSUPPORTED_TYPES:
    GET_STATE_DISPATCH_FUNCTIONS.append((type_, unsupported_get_state))

# tuples of type and function that creates the instance of that type
NODE_TYPE_MAPPING = {
    ("SGDNode", PROTOCOL): SGDNode,
    ("TreeNode", PROTOCOL): TreeNode,
}

# TODO: remove once support for sklearn<1.2 is dropped.
# Starting from sklearn 1.2, _DictWithDeprecatedKeys is removed as it's no
# longer needed for GraphicalLassoCV, see #187.
if _DictWithDeprecatedKeys is not None:
    GET_STATE_DISPATCH_FUNCTIONS.append(
        (_DictWithDeprecatedKeys, _DictWithDeprecatedKeys_get_state)
    )
    NODE_TYPE_MAPPING[
        ("_DictWithDeprecatedKeysNode", PROTOCOL)
    ] = _DictWithDeprecatedKeysNode  # type: ignore



import difflib
import json
import os
import re
from pathlib import Path
from unittest.mock import Mock, patch

import pytest
import yaml  # type: ignore

from skops.card import parse_modelcard
from skops.card._parser import PandocParser, check_pandoc_installed

try:
    check_pandoc_installed()
except FileNotFoundError:
    # not installed, skip
    pytest.skip(reason="These tests require a recent pandoc", allow_module_level=True)


EXAMPLE_CARDS = [
    # actual model cards from HF hub
    "bert-base-uncased.md",
    "clip-vit-large-patch14.md",
    "gpt2.md",
    "specter.md",
    "vit-base-patch32-224-in21k.md",
    # not a model card
    "toy-example.md",
]


def _assert_meta_equal(meta0, meta1):
    # we cannot guarantee the order of metadata items, so we compare parsed
    # dicts, but not strings directly
    assert yaml.safe_load("".join(meta0)) == yaml.safe_load("".join(meta1))


def assert_readme_files_almost_equal(file0, file1, diff):
    """Check that the two model cards are identical, but allow differences as
    defined in the ``diff`` file

    The metainfo is compared separately, as the order of the items is not
    guaranteed to be stable.
    """
    with open(file0, "r") as f:
        readme0 = f.readlines()

    with open(file1, "r") as f:
        readme1 = f.readlines()

    sep = "---\n"
    # we look for 2nd occurrence, so skip first char to not match 1st occurrence
    if sep in readme0[1:]:  # only check if metainfo is present
        idx0, idx1 = readme0[1:].index(sep) + 1, readme1[1:].index(sep) + 1
        meta0, meta1 = readme0[1:idx0], readme1[1:idx1]
        readme0, readme1 = readme0[idx0:], readme1[idx1:]
        _assert_meta_equal(meta0, meta1)

    # exclude trivial case of both being empty
    assert readme0
    assert readme1

    diff_actual = list(difflib.unified_diff(readme0, readme1, n=0))

    with open(diff, "r") as f:
        diff_expected = f.readlines()

    assert diff_actual == diff_expected


@pytest.mark.parametrize("file_name", EXAMPLE_CARDS, ids=EXAMPLE_CARDS)
def test_example_model_cards(tmp_path, file_name):
    """Test that the difference between original and parsed model card is
    acceptable

    For this test, model cards for some of the most popular models on HF Hub
    were retrieved and stored in the ./examples folder. This test checks that
    these model cards can be successfully parsed and that the output is *almost*
    the same.

    We don't expect the output to be 100% identical, see the limitations listed
    in ``parse_modelcard``. Instead, we assert that the diff corresponds to the
    expected diff, which is also checked in.

    So e.g. for "specter.md", we expect that the diff will be the same diff as
    in "specter.md.diff".

    """
    path = Path(os.getcwd()) / "skops" / "card" / "tests" / "examples"
    file0 = path / file_name
    diff = (path / file_name).with_suffix(".md.diff")
    parsed_card = parse_modelcard(file0)
    file1 = tmp_path / "readme-parsed.md"
    parsed_card.save(file1)

    assert_readme_files_almost_equal(file0, file1, diff)


def test_unknown_pandoc_item_raises():
    source = json.dumps(
        {
            "pandoc-api-version": [1, 22, 2, 1],
            "meta": {},
            "blocks": [
                {
                    "t": "Header",
                    "c": [1, ["section", [], []], [{"t": "Str", "c": "section"}]],
                },
                {"c": "valid", "t": "Str"},
                {"t": "does-not-exist", "c": []},
                {"c": "okay", "t": "Str"},
            ],
        }
    )
    parser = PandocParser(source)
    msg = (
        "The parsed document contains 'does-not-exist', which is not "
        "supported yet, please open an issue on GitHub"
    )
    with pytest.raises(ValueError, match=re.escape(msg)):
        parser.generate()


def test_content_without_section_raises():
    source = json.dumps(
        {
            "pandoc-api-version": [1, 22, 2, 1],
            "meta": {},
            "blocks": [
                {"c": "whoops", "t": "Str"},
            ],
        }
    )
    parser = PandocParser(source)
    msg = (
        "Trying to add content but there is no current section, this is probably a "
        "bug, please open an issue on GitHub"
    )
    with pytest.raises(ValueError, match=re.escape(msg)):
        parser.generate()


def test_unsupported_markup_raises():
    match = re.escape("Markup of type does-not-exist is not supported (yet)")
    with pytest.raises(ValueError, match=match):
        PandocParser(source="", markup_type="does-not-exist")


def test_check_pandoc_installed_no_min_version_works():
    # check that it doesn't raise
    check_pandoc_installed(min_version=None)


def test_check_pandoc_installed_min_version_too_high_raises():
    match = re.escape("Pandoc version too low, expected at least 999.9.9, got")
    with pytest.raises(ValueError, match=match):
        check_pandoc_installed(min_version="999.9.9")


def test_pandoc_not_installed():
    def raise_filenotfound(*args, **kwargs):
        # error raised when trying to run subprocess on non-existing command
        raise FileNotFoundError("[Errno 2] No such file or directory: 'pandoc'")

    with patch("subprocess.run", raise_filenotfound):
        match = re.escape(
            "This feature requires the pandoc library to be installed on your system"
        )
        with pytest.raises(FileNotFoundError, match=match):
            check_pandoc_installed()


def test_pandoc_version_cannot_be_determined():
    mock = Mock()
    with patch("subprocess.run", mock):
        match = re.escape("Could not determine version of pandoc")
        with pytest.raises(RuntimeError, match=match):
            check_pandoc_installed()



from __future__ import annotations

import importlib
import sys
from dataclasses import dataclass, field
from functools import singledispatch
from typing import Any, Type
from zipfile import ZipFile

from ._protocol import PROTOCOL


# The following two functions are copied from cpython's pickle.py file.
# ---------------------------------------------------------------------
def _getattribute(obj, name):
    for subpath in name.split("."):
        if subpath == "<locals>":
            raise AttributeError(
                "Can't get local attribute {!r} on {!r}".format(name, obj)
            )
        try:
            parent = obj
            obj = getattr(obj, subpath)
        except AttributeError:
            raise AttributeError(
                "Can't get attribute {!r} on {!r}".format(name, obj)
            ) from None
    return obj, parent


# This function is particularly used to detect the path of functions such as
# ufuncs. It returns the full path, instead of returning the module name.
def whichmodule(obj: Any, name: str) -> str:
    """Find the module an object belong to."""
    module_name = getattr(obj, "__module__", None)
    if module_name is not None:
        return module_name
    # Protect the iteration by using a list copy of sys.modules against dynamic
    # modules that trigger imports of other modules upon calls to getattr.
    for module_name, module in sys.modules.copy().items():
        if (
            module_name == "__main__"
            or module_name == "__mp_main__"  # bpo-42406
            or module is None
        ):
            continue
        try:
            if _getattribute(module, name)[0] is obj:
                return module_name
        except AttributeError:
            pass
    return "__main__"


# ---------------------------------------------------------------------


def _import_obj(module: str, cls_or_func: str, package: str | None = None) -> Any:
    return getattr(importlib.import_module(module, package=package), cls_or_func)


def gettype(module_name: str, cls_or_func: str) -> Type[Any]:
    if module_name and cls_or_func:
        return _import_obj(module_name, cls_or_func)

    raise ValueError(f"Object {cls_or_func} of module {module_name} is unknown")


def get_module(obj: Any) -> str:
    """Find module for given object

    If the module cannot be identified, it's assumed to be "__main__".

    Parameters
    ----------
    obj: Any
       Object whose module is requested.

    Returns
    -------
    name: str
        Name of the module.

    """
    return whichmodule(obj, obj.__name__)


@dataclass(frozen=True)
class SaveContext:
    """Context required for saving the objects

    This context is passed to each ``get_state_*`` function.

    Parameters
    ----------
    zip_file: zipfile.ZipFile
        The zip file to write the data to, must be in write mode.

    protocol: int
        The protocol of the persistence format. Right now, there is only
        protocol 0, but this leaves the door open for future changes.

    """

    zip_file: ZipFile
    protocol: int = PROTOCOL
    memo: dict[int, Any] = field(default_factory=dict)

    def memoize(self, obj: Any) -> int:
        # Currently, the only purpose for saving the object id is to make sure
        # that for the length of the context that the main object is being
        # saved, all attributes persist, so that the same id cannot be re-used
        # for different objects.
        obj_id = id(obj)
        if obj_id not in self.memo:
            self.memo[obj_id] = obj
        return obj_id

    def clear_memo(self) -> None:
        self.memo.clear()


@dataclass(frozen=True)
class LoadContext:
    """Context required for loading an object

    This context is passed to each ``*Node`` class when loading an object.

    Parameters
    ----------
    src: zipfile.ZipFile
        The zip file the target object is saved in
    """

    src: ZipFile
    protocol: int
    memo: dict[int, Any] = field(default_factory=dict)

    def memoize(self, obj: Any, id: int) -> None:
        self.memo[id] = obj

    def get_object(self, id: int) -> Any:
        return self.memo.get(id)


@singledispatch
def _get_state(obj, save_context: SaveContext):
    # This function should never be called directly. Instead, it is used to
    # dispatch to the correct implementation of get_state for the given type of
    # its first argument.
    raise TypeError(f"Getting the state of type {type(obj)} is not supported yet")


def get_state(value, save_context: SaveContext) -> dict[str, Any]:
    # This is a helper function to try to get the state of an object. If it
    # fails with `get_state`, we try with json.dumps, if that fails, we raise
    # the original error alongside the json error.

    # TODO: This should help with fixing recursive references.
    # if id(value) in save_context.memo:
    #     return {
    #         "__module__": None,
    #         "__class__": None,
    #         "__id__": id(value),
    #         "__loader__": "CachedNode",
    #     }

    __id__ = save_context.memoize(obj=value)

    res = _get_state(value, save_context)

    res["__id__"] = __id__
    return res


def get_type_name(t: Any) -> str:
    """Helper function to take in a type, and return its name as a string"""
    return f"{get_module(t)}.{t.__name__}"


def get_type_paths(types: Any) -> list[str]:
    """Helper function that takes in a types,
    and converts any the types found to a list of strings.

    Parameters
    ----------
    types: Any
        Types to get. Can be either a string, a single type, or a list of strings
        and types.

    Returns
    ----------
    types_list: list of str
        The list of types, all as strings, e.g. ``["builtins.list"]``.

    """
    if not types:
        return []
    if not isinstance(types, (list, tuple)):
        types = [types]

    return [get_type_name(t) if not isinstance(t, str) else t for t in types]



import importlib
import inspect
import io
import json
import operator
import sys
import warnings
from collections import Counter
from functools import partial, wraps
from pathlib import Path
from zipfile import ZipFile

import joblib
import numpy as np
import pytest
from scipy import sparse, special
from sklearn.base import BaseEstimator, is_regressor
from sklearn.compose import ColumnTransformer
from sklearn.datasets import load_sample_images, make_classification, make_regression
from sklearn.decomposition import SparseCoder
from sklearn.exceptions import SkipTestWarning
from sklearn.experimental import enable_halving_search_cv  # noqa
from sklearn.linear_model import LogisticRegression
from sklearn.model_selection import (
    GridSearchCV,
    GroupKFold,
    HalvingGridSearchCV,
    HalvingRandomSearchCV,
    KFold,
    RandomizedSearchCV,
    ShuffleSplit,
    check_cv,
)
from sklearn.neighbors import KNeighborsClassifier, KNeighborsRegressor
from sklearn.pipeline import FeatureUnion, Pipeline
from sklearn.preprocessing import (
    FunctionTransformer,
    MinMaxScaler,
    Normalizer,
    PolynomialFeatures,
    StandardScaler,
)
from sklearn.utils import all_estimators, check_random_state
from sklearn.utils._tags import _safe_tags
from sklearn.utils._testing import SkipTest, set_random_state
from sklearn.utils.estimator_checks import (
    _construct_instance,
    _enforce_estimator_tags_y,
    _get_check_estimator_ids,
)

import skops
from skops.io import dump, dumps, get_untrusted_types, load, loads
from skops.io._audit import NODE_TYPE_MAPPING, get_tree
from skops.io._sklearn import UNSUPPORTED_TYPES
from skops.io._trusted_types import SCIPY_UFUNC_TYPE_NAMES, SKLEARN_ESTIMATOR_TYPE_NAMES
from skops.io._utils import LoadContext, SaveContext, _get_state, get_state, gettype
from skops.io.exceptions import UnsupportedTypeException, UntrustedTypesFoundException
from skops.io.tests._utils import assert_method_outputs_equal, assert_params_equal

# Default settings for X
N_SAMPLES = 50
N_FEATURES = 20


@pytest.fixture(autouse=True, scope="module")
def debug_dispatch_functions():
    # Patch the get_state and get_tree methods to add some sanity checks on
    # them. Specifically, we test that the arguments of the functions all follow
    # the same pattern to enforce consistency and that the "state" is either a
    # dict with specified keys or a primitive type.

    def debug_get_state(func):
        # Check consistency of argument names, output type, and that the output,
        # if a dict, has certain keys, or if not a dict, is a primitive type.
        signature = inspect.signature(func)
        assert list(signature.parameters.keys()) == ["obj", "save_context"]

        @wraps(func)
        def wrapper(obj, save_context):
            # NB: __id__ set in main 'get_state' func, so no check here
            result = func(obj, save_context)

            assert "__class__" in result
            assert "__module__" in result
            assert "__loader__" in result

            return result

        return wrapper

    def debug_get_tree(func):
        # check consistency of argument names and input type
        signature = inspect.signature(func)
        assert list(signature.parameters.keys()) == ["state", "load_context", "trusted"]

        @wraps(func)
        def wrapper(state, load_context, trusted):
            assert "__class__" in state
            assert "__module__" in state
            assert "__loader__" in state
            assert "__id__" in state
            assert isinstance(load_context, LoadContext)

            result = func(state, load_context, trusted)
            return result

        return wrapper

    modules = ["._general", "._numpy", "._scipy", "._sklearn"]
    for module_name in modules:
        # overwrite exposed functions for get_state and get_tree
        module = importlib.import_module(module_name, package="skops.io")
        for cls, method in getattr(module, "GET_STATE_DISPATCH_FUNCTIONS", []):
            _get_state.register(cls)(debug_get_state(method))
        for key, method in NODE_TYPE_MAPPING.copy().items():
            NODE_TYPE_MAPPING[key] = debug_get_tree(method)


def _tested_estimators(type_filter=None):
    for name, Estimator in all_estimators(type_filter=type_filter):
        if Estimator in UNSUPPORTED_TYPES:
            continue
        try:
            # suppress warnings here for skipped estimators.
            with warnings.catch_warnings():
                warnings.filterwarnings(
                    "ignore",
                    category=SkipTestWarning,
                    message="Can't instantiate estimator",
                )
                estimator = _construct_instance(Estimator)
                # with the kind of data we pass, it needs to be 1 for the few
                # estimators which have this.
                if "n_components" in estimator.get_params():
                    estimator.set_params(n_components=1)
                    # Then n_best needs to be <= n_components
                    if "n_best" in estimator.get_params():
                        estimator.set_params(n_best=1)
                if "patch_size" in estimator.get_params():
                    # set patch size to fix PatchExtractor test.
                    estimator.set_params(patch_size=(3, 3))
        except SkipTest:
            continue

        yield estimator

    # nested Pipeline & FeatureUnion
    # fmt: off
    yield Pipeline([
        ("features", FeatureUnion([
            ("scaler", StandardScaler()),
            ("scaled-poly", Pipeline([
                ("polys", FeatureUnion([
                    ("poly1", PolynomialFeatures()),
                    ("poly2", PolynomialFeatures(degree=3, include_bias=False))
                ])),
                ("scale", MinMaxScaler()),
            ])),
        ])),
        ("clf", LogisticRegression(random_state=0, solver="liblinear")),
    ])
    # fmt: on

    # FunctionTransformer with numpy functions
    yield FunctionTransformer(
        func=np.sqrt,
        inverse_func=np.square,
    )

    # FunctionTransformer with scipy functions - problem is that they look like
    # numpy ufuncs
    yield FunctionTransformer(
        func=special.erf,
        inverse_func=special.erfinv,
    )

    # partial functions should be supported
    yield FunctionTransformer(
        func=partial(np.add, 10),
        inverse_func=partial(np.add, -10),
    )

    yield KNeighborsClassifier(algorithm="kd_tree")
    yield KNeighborsRegressor(algorithm="ball_tree")

    yield ColumnTransformer(
        [
            ("norm1", Normalizer(norm="l1"), [0]),
            ("norm2", Normalizer(norm="l1"), [1, 2]),
            ("norm3", Normalizer(norm="l1"), [True] + (N_FEATURES - 1) * [False]),
            ("norm4", Normalizer(norm="l1"), np.array([1, 2])),
            ("norm5", Normalizer(norm="l1"), slice(3)),
            ("norm6", Normalizer(norm="l1"), slice(-10, -3, 2)),
        ],
    )

    yield GridSearchCV(
        LogisticRegression(random_state=0, solver="liblinear"),
        {"C": [1, 2, 3, 4, 5]},
    )

    yield HalvingGridSearchCV(
        LogisticRegression(random_state=0, solver="liblinear"),
        {"C": [1, 2, 3, 4, 5]},
    )

    yield HalvingRandomSearchCV(
        LogisticRegression(random_state=0, solver="liblinear"),
        {"C": [1, 2, 3, 4, 5]},
    )

    yield RandomizedSearchCV(
        LogisticRegression(random_state=0, solver="liblinear"),
        {"C": [1, 2, 3, 4, 5]},
        n_iter=3,
    )

    dictionary = np.random.randint(-2, 3, size=(5, N_FEATURES)).astype(float)
    yield SparseCoder(
        dictionary=dictionary,
        transform_algorithm="lasso_lars",
    )


def _tested_ufuncs():
    for full_name in SCIPY_UFUNC_TYPE_NAMES:
        module_name, _, ufunc_name = full_name.rpartition(".")
        yield gettype(module_name=module_name, cls_or_func=ufunc_name)


def _unsupported_estimators(type_filter=None):
    for name, Estimator in all_estimators(type_filter=type_filter):
        if Estimator not in UNSUPPORTED_TYPES:
            continue
        try:
            # suppress warnings here for skipped estimators.
            with warnings.catch_warnings():
                warnings.filterwarnings(
                    "ignore",
                    category=SkipTestWarning,
                    message="Can't instantiate estimator",
                )
                estimator = _construct_instance(Estimator)
                # with the kind of data we pass, it needs to be 1 for the few
                # estimators which have this.
                if "n_components" in estimator.get_params():
                    estimator.set_params(n_components=1)
                    # Then n_best needs to be <= n_components
                    if "n_best" in estimator.get_params():
                        estimator.set_params(n_best=1)
                if "patch_size" in estimator.get_params():
                    # set patch size to fix PatchExtractor test.
                    estimator.set_params(patch_size=(3, 3))
        except SkipTest:
            continue

        yield estimator


@pytest.mark.parametrize(
    "estimator", _tested_estimators(), ids=_get_check_estimator_ids
)
def test_can_persist_non_fitted(estimator):
    """Check that non-fitted estimators can be persisted."""
    loaded = loads(dumps(estimator), trusted=True)
    assert_params_equal(estimator.get_params(), loaded.get_params())


def get_input(estimator):
    # Return a valid input for estimator.fit

    # TODO: make this a parameter and test with sparse data
    # TODO: try with pandas.DataFrame as well
    if is_regressor(estimator):
        # classifier data can lead to failure of certain regressors to fit, e.g.
        # RANSAC in sklearn 0.24, so regression data is needed
        X, y = make_regression(
            n_samples=N_SAMPLES, n_features=N_FEATURES, random_state=0
        )
    else:
        X, y = make_classification(
            n_samples=N_SAMPLES, n_features=N_FEATURES, random_state=0
        )
    y = _enforce_estimator_tags_y(estimator, y)
    tags = _safe_tags(estimator)

    if tags["pairwise"] is True:
        return np.random.rand(N_FEATURES, N_FEATURES), None

    if "2darray" in tags["X_types"]:
        # Some models require positive X
        return np.abs(X), y

    if "1darray" in tags["X_types"]:
        return X[:, 0], y

    if "3darray" in tags["X_types"]:
        return load_sample_images().images[1], None

    if "1dlabels" in tags["X_types"]:
        # model only expects y
        return y, None

    if "2dlabels" in tags["X_types"]:
        return [(1, 2), (3,)], None

    if "categorical" in tags["X_types"]:
        X = [["Male", 1], ["Female", 3], ["Female", 2]]
        y = y[: len(X)] if tags["requires_y"] else None
        return X, y

    if "dict" in tags["X_types"]:
        return [{"foo": 1, "bar": 2}, {"foo": 3, "baz": 1}], None

    if "string" in tags["X_types"]:
        return [
            "This is the first document.",
            "This document is the second document.",
            "And this is the third one.",
            "Is this the first document?",
        ], None

    if tags["X_types"] == "sparse":
        # TfidfTransformer in sklearn 0.24 needs this
        return sparse.csr_matrix(X), y

    raise ValueError(f"Unsupported X type for estimator: {tags['X_types']}")


@pytest.mark.parametrize(
    "estimator", _tested_estimators(), ids=_get_check_estimator_ids
)
def test_can_persist_fitted(estimator):
    """Check that fitted estimators can be persisted and return the right results."""
    set_random_state(estimator, random_state=0)

    X, y = get_input(estimator)
    tags = _safe_tags(estimator)
    if tags.get("requires_fit", True):
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", module="sklearn")
            if y is not None:
                estimator.fit(X, y)
            else:
                estimator.fit(X)

    # test that we can get a list of untrusted types. This is a smoke test
    # to make sure there are no errors running this method.
    # it is in this test to save time, as it requires a fitted estimator.
    dumped = dumps(estimator)
    untrusted_types = get_untrusted_types(data=dumped)

    loaded = loads(dumped, trusted=untrusted_types)
    assert_params_equal(estimator.__dict__, loaded.__dict__)

    assert not any(type_ in SKLEARN_ESTIMATOR_TYPE_NAMES for type_ in untrusted_types)
    assert not any(type_ in SCIPY_UFUNC_TYPE_NAMES for type_ in untrusted_types)
    assert_method_outputs_equal(estimator, loaded, X)


@pytest.mark.parametrize("ufunc", _tested_ufuncs(), ids=SCIPY_UFUNC_TYPE_NAMES)
def test_can_trust_ufuncs(ufunc):
    dumped = dumps(ufunc)
    untrusted_types = get_untrusted_types(data=dumped)
    assert len(untrusted_types) == 0
    # TODO: extend with numpy ufuncs


@pytest.mark.parametrize(
    "estimator", _unsupported_estimators(), ids=_get_check_estimator_ids
)
def test_unsupported_type_raises(estimator):
    """Estimators that are known to fail should raise an error"""
    set_random_state(estimator, random_state=0)

    X, y = get_input(estimator)
    tags = _safe_tags(estimator)
    if tags.get("requires_fit", True):
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", module="sklearn")
            if y is not None:
                estimator.fit(X, y)
            else:
                estimator.fit(X)

    msg = f"Objects of type {estimator.__class__.__name__} are not supported yet"
    with pytest.raises(UnsupportedTypeException, match=msg):
        dumps(estimator)


class RandomStateEstimator(BaseEstimator):
    def __init__(self, random_state=None):
        self.random_state = random_state

    def fit(self, X, y, **fit_params):
        if isinstance(self.random_state, np.random.Generator):
            # forwards compatibility with np.random.Generator
            self.random_state_ = self.random_state
        else:
            self.random_state_ = check_random_state(self.random_state)
        return self


@pytest.mark.parametrize(
    "random_state",
    [
        None,
        0,
        np.random.RandomState(42),
        np.random.default_rng(),
        np.random.Generator(np.random.PCG64DXSM(seed=123)),
    ],
    ids=["None", "int", "RandomState", "default_rng", "Generator"],
)
def test_random_state(random_state):
    # Numpy random Generators
    # (https://numpy.org/doc/stable/reference/random/generator.html) are not
    # supported by sklearn yet but will be in the future, thus they're tested
    # here
    est = RandomStateEstimator(random_state=random_state).fit(None, None)
    est.random_state_.random(123)  # move RNG forwards

    dumped = dumps(est)
    untrusted_types = get_untrusted_types(data=dumped)
    loaded = loads(dumped, trusted=untrusted_types)

    if hasattr(est, "__dict__"):
        # what to do if object has no __dict__, like Generator?
        assert_params_equal(est.__dict__, loaded.__dict__)

    rand_floats_expected = est.random_state_.random(100)
    rand_floats_loaded = loaded.random_state_.random(100)
    np.testing.assert_equal(rand_floats_loaded, rand_floats_expected)


class CVEstimator(BaseEstimator):
    def __init__(self, cv=None):
        self.cv = cv

    def fit(self, X, y, **fit_params):
        self.cv_ = check_cv(self.cv)
        return self

    def split(self, X, **kwargs):
        return list(self.cv_.split(X, **kwargs))


@pytest.mark.parametrize(
    "cv",
    [
        None,
        3,
        KFold(4, shuffle=True, random_state=42),
        GroupKFold(5),
        ShuffleSplit(6, random_state=np.random.RandomState(123)),
    ],
)
def test_cross_validator(cv):
    est = CVEstimator(cv=cv).fit(None, None)
    loaded = loads(dumps(est), trusted=True)
    X, y = make_classification(
        n_samples=N_SAMPLES, n_features=N_FEATURES, random_state=0
    )

    kwargs = {}
    name = est.cv_.__class__.__name__.lower()
    if "stratified" in name:
        kwargs["y"] = y
    if "group" in name:
        kwargs["groups"] = np.random.randint(0, 5, size=len(y))

    splits_est = est.split(X, **kwargs)
    splits_loaded = loaded.split(X, **kwargs)
    assert len(splits_est) == len(splits_loaded)
    for split_est, split_loaded in zip(splits_est, splits_loaded):
        np.testing.assert_equal(split_est, split_loaded)


class EstimatorWith2dObjectArray(BaseEstimator):
    def fit(self, X, y=None, **fit_params):
        self.obj_array_ = np.array([[1, "2"], [3.0, None]])
        return self


@pytest.mark.parametrize(
    "transpose",
    [
        False,
        pytest.param(True, marks=pytest.mark.xfail(raises=AssertionError)),
    ],
)
def test_numpy_object_dtype_2d_array(transpose):
    # Explicitly test multi-dimensional (i.e. more than 1) object arrays, since
    # those use json instead of numpy.save/load and some errors may only occur
    # with multi-dimensional arrays (e.g. mismatched contiguity). For
    # F-contiguous object arrays, this test currently fails, as the array is
    # loaded as C-contiguous.
    est = EstimatorWith2dObjectArray().fit(None)
    if transpose:
        est.obj_array_ = est.obj_array_.T

    loaded = loads(dumps(est), trusted=True)
    assert_params_equal(est.__dict__, loaded.__dict__)


def test_metainfo():
    class MyEstimator(BaseEstimator):
        """Estimator with attributes of different supported types"""

        def fit(self, X, y=None, **fit_params):
            self.builtin_ = [1, 2, 3]
            self.stdlib_ = Counter([10, 20, 20, 30, 30, 30])
            self.numpy_ = np.arange(5)
            self.sparse_ = sparse.csr_matrix([[0, 1], [1, 0]])
            self.sklearn_ = LogisticRegression()
            # create a nested data structure to check if that works too
            self.nested_ = {
                "builtin_": self.builtin_,
                "stdlib_": self.stdlib_,
                "numpy_": self.numpy_,
                "sparse_": self.sparse_,
                "sklearn_": self.sklearn_,
            }
            return self

    # safe and load the schema
    estimator = MyEstimator().fit(None)
    dumped = dumps(estimator)
    schema = json.loads(ZipFile(io.BytesIO(dumped)).read("schema.json"))

    # check some schema metainfo
    assert schema["protocol"] == skops.io._protocol.PROTOCOL
    assert schema["_skops_version"] == skops.__version__

    # additionally, check following metainfo: class, module, and version
    expected = {
        "builtin_": {
            "__class__": "list",
            "__module__": "builtins",
        },
        "stdlib_": {
            "__class__": "Counter",
            "__module__": "collections",
        },
        "numpy_": {
            "__class__": "ndarray",
            "__module__": "numpy",
        },
        "sparse_": {
            "__class__": "csr_matrix",
            "__module__": "scipy.sparse",
        },
        "sklearn_": {
            "__class__": "LogisticRegression",
            "__module__": "sklearn.linear_model",
        },
    }
    # check both the top level state and the nested state
    states = (
        schema["content"]["content"],
        schema["content"]["content"]["nested_"]["content"],
    )
    for key, val_expected in expected.items():
        for state in states:
            val_state = state[key]
            # check presence of "content"/"file" but not exact values
            assert ("content" in val_state) or ("file" in val_state)
            assert val_state["__class__"] == val_expected["__class__"]
            # We don't want to compare full module structures, because they can
            # change across versions, e.g. 'scipy.sparse.csr' moving to
            # 'scipy.sparse._csr'.
            assert val_state["__module__"].startswith(val_expected["__module__"])


class EstimatorIdenticalArrays(BaseEstimator):
    """Estimator that stores multiple references to the same array"""

    def fit(self, X, y=None, **fit_params):
        # each block below should reference the same file
        self.X = X
        self.X_2 = X
        self.X_list = [X, X]
        self.X_dict = {"a": X, 2: X}

        # copies are not deduplicated
        X_copy = X.copy()
        self.X_copy = X_copy
        self.X_copy2 = X_copy

        # transposed matrices are not the same
        X_T = X.T
        self.X_T = X_T
        self.X_T2 = X_T

        # slices are not the same
        self.vector = X[0]

        self.vector_2 = X[0]

        self.scalar = X[0, 0]

        self.scalar_2 = X[0, 0]

        # deduplication should work on sparse matrices
        X_sparse = sparse.csr_matrix(X)
        self.X_sparse = X_sparse
        self.X_sparse2 = X_sparse

        return self


def test_identical_numpy_arrays_not_duplicated():
    # Test that identical numpy arrays are not stored multiple times
    X = np.random.random((10, 5))
    estimator = EstimatorIdenticalArrays().fit(X)
    dumped = dumps(estimator)
    loaded = loads(dumped, trusted=True)
    assert_params_equal(estimator.__dict__, loaded.__dict__)

    # check number of numpy arrays stored on disk
    with ZipFile(io.BytesIO(dumped), "r") as input_zip:
        files = input_zip.namelist()
    # expected number of files are:
    # schema, X, X_copy, X_t, 2 vectors, 2 scalars, X_sparse = 9
    expected_files = 9
    num_files = len(files)
    assert num_files == expected_files


class NumpyDtypeObjectEstimator(BaseEstimator):
    """An estimator with a numpy array of dtype object"""

    def fit(self, X, y=None, **fit_params):
        self.obj_ = np.zeros(3, dtype=object)
        return self


def test_numpy_dtype_object_does_not_store_broken_file():
    # This addresses a specific bug where trying to store an object numpy array
    # resulted in the creation of a broken .npy file being left over. This is
    # because numpy tries to write to the file until it encounters an error and
    # raises, but then doesn't clean up said file. Before the bugfix in #150, we
    # would include that broken file in the zip archive, although we wouldn't do
    # anything with it. Here we test that no such file exists.
    estimator = NumpyDtypeObjectEstimator().fit(None)
    dumped = dumps(estimator)
    with ZipFile(io.BytesIO(dumped), "r") as input_zip:
        files = input_zip.namelist()

    # this estimator should not have any numpy file
    assert not any(file.endswith(".npy") for file in files)


def test_loads_from_str():
    # loads expects bytes, not str
    msg = "Can't load skops format from string, pass bytes"
    with pytest.raises(TypeError, match=msg):
        loads("this is a string")


def test_get_tree_unknown_type_error_msg():
    state = get_state(("hi", [123]), SaveContext(None))
    state["__loader__"] = "this_get_tree_does_not_exist"
    msg = "Can't find loader this_get_tree_does_not_exist for type builtins.tuple."
    with pytest.raises(TypeError, match=msg):
        get_tree(state, LoadContext(None, -1))


class _BoundMethodHolder:
    """Used to test the ability to serialize and deserialize bound methods"""

    def __init__(self, object_state: str):
        # Initialize with some state to make sure state is persisted
        self.object_state = object_state
        # bind some method to this object, could be any persistable function
        self.chosen_function = np.log

    def bound_method(self, x):
        return self.chosen_function(x)

    def other_bound_method(self, x):
        # arbitrary other function, used for checking single instance loaded
        return self.chosen_function(x)


class TestPersistingBoundMethods:
    @staticmethod
    def assert_transformer_persisted_correctly(
        loaded_transformer: FunctionTransformer,
        original_transformer: FunctionTransformer,
    ):
        """Checks that a persisted and original transformer are equivalent, including
        the func passed to it
        """
        assert loaded_transformer.func.__name__ == original_transformer.func.__name__

        assert_params_equal(
            loaded_transformer.func.__self__.__dict__,
            original_transformer.func.__self__.__dict__,
        )
        assert_params_equal(loaded_transformer.__dict__, original_transformer.__dict__)

    @staticmethod
    def assert_bound_method_holder_persisted_correctly(
        original_obj: _BoundMethodHolder, loaded_obj: _BoundMethodHolder
    ):
        """Checks that the persisted and original instances of _BoundMethodHolder are
        equivalent
        """
        assert original_obj.bound_method.__name__ == loaded_obj.bound_method.__name__
        assert original_obj.chosen_function == loaded_obj.chosen_function

        assert_params_equal(original_obj.__dict__, loaded_obj.__dict__)

    def test_for_base_case_returns_as_expected(self):
        initial_state = "This is an arbitrary state"
        obj = _BoundMethodHolder(object_state=initial_state)
        bound_function = obj.bound_method
        transformer = FunctionTransformer(func=bound_function)

        loaded_transformer = loads(dumps(transformer), trusted=True)
        loaded_obj = loaded_transformer.func.__self__

        self.assert_transformer_persisted_correctly(loaded_transformer, transformer)
        self.assert_bound_method_holder_persisted_correctly(obj, loaded_obj)

    def test_when_object_is_changed_after_init_works_as_expected(self):
        # given change to object with bound method after initialisation,
        # make sure still persists correctly

        initial_state = "This is an arbitrary state"
        obj = _BoundMethodHolder(object_state=initial_state)
        obj.chosen_function = np.sqrt
        bound_function = obj.bound_method

        transformer = FunctionTransformer(func=bound_function)

        loaded_transformer = loads(dumps(transformer), trusted=True)
        loaded_obj = loaded_transformer.func.__self__

        self.assert_transformer_persisted_correctly(loaded_transformer, transformer)
        self.assert_bound_method_holder_persisted_correctly(obj, loaded_obj)

    def test_works_when_given_multiple_bound_methods_attached_to_single_instance(self):
        obj = _BoundMethodHolder(object_state="")

        transformer = FunctionTransformer(
            func=obj.bound_method, inverse_func=obj.other_bound_method
        )

        loaded_transformer = loads(dumps(transformer), trusted=True)

        # check that both func and inverse_func are from the same object instance
        loaded_0 = loaded_transformer.func.__self__
        loaded_1 = loaded_transformer.inverse_func.__self__
        assert loaded_0 is loaded_1

    @pytest.mark.xfail(reason="Failing due to circular self reference")
    def test_scipy_stats(self, tmp_path):
        from scipy import stats

        estimator = FunctionTransformer(func=stats.zipf)
        loads(dumps(estimator), trusted=True)


class CustomEstimator(BaseEstimator):
    """Estimator with np array, np scalar, and sparse matrix attribute"""

    def fit(self, X, y=None):
        self.numpy_array = np.zeros(3)
        self.numpy_scalar = np.ones(1)[0]
        self.sparse_matrix = sparse.csr_matrix(np.arange(3))
        return self


def test_dump_to_and_load_from_disk(tmp_path):
    # Test saving to and loading from disk. Functionality-wise, this is almost
    # identical to saving to and loading from memory using dumps and loads.
    # Therefore, only test functionality that is specific to dump and load.

    estimator = CustomEstimator().fit(None)
    f_name = tmp_path / "estimator.skops"
    dump(estimator, f_name)
    file = Path(f_name)
    assert file.exists()

    with ZipFile(f_name, "r") as input_zip:
        files = input_zip.namelist()

    # there should be 4 files in total, schema.json, 2 np arrays, and 1 sparse matrix
    assert len(files) == 4
    assert "schema.json" in files

    num_array_files = sum(1 for file in files if file.endswith(".npy"))
    assert num_array_files == 2
    num_sparse_files = sum(1 for file in files if file.endswith(".npz"))
    assert num_sparse_files == 1

    # check that schema is valid json by loading it
    json.loads(ZipFile(f_name).read("schema.json"))

    # load and compare the actual estimator
    loaded = load(f_name, trusted=True)
    assert_params_equal(loaded.__dict__, estimator.__dict__)


def test_disk_and_memory_are_identical(tmp_path):
    # Test that model hashes are the same for models stored on disk and in
    # memory.
    # Use a somewhat complex model.
    # fmt: off
    estimator = Pipeline([
        ("features", FeatureUnion([
            ("scaler", StandardScaler()),
            ("scaled-poly", Pipeline([
                ("polys", FeatureUnion([
                    ("poly1", PolynomialFeatures()),
                    ("poly2", PolynomialFeatures(degree=3, include_bias=False))
                ])),
                ("square-root", FunctionTransformer(np.sqrt)),
                ("scale", MinMaxScaler()),
            ])),
        ])),
        ("clf", LogisticRegression(random_state=0, solver="liblinear")),
    ]).fit([[0, 1], [2, 3], [4, 5]], [0, 1, 2])
    # fmt: on

    f_name = tmp_path / "estimator.skops"
    dump(estimator, f_name)
    loaded_disk = load(f_name, trusted=True)
    loaded_memory = loads(dumps(estimator), trusted=True)

    assert joblib.hash(loaded_disk) == joblib.hash(loaded_memory)


def test_dump_and_load_with_file_wrapper(tmp_path):
    # The idea here is to make it possible to use dump and load with a file
    # wrapper, i.e. using 'with open(...)'. This makes it easier to search and
    # replace pickle dump and load by skops dump and load.
    estimator = LogisticRegression().fit([[0, 1], [2, 3], [4, 5]], [0, 1, 1])
    f_name = tmp_path / "estimator.skops"

    with open(f_name, "wb") as f:
        dump(estimator, f)
    with open(f_name, "rb") as f:
        loaded = load(f, trusted=True)

    assert_params_equal(loaded.__dict__, estimator.__dict__)


@pytest.mark.parametrize(
    "obj",
    [
        np.array([1, 2]),
        [1, 2, 3],
        {1: 1, 2: 2},
        {1, 2, 3},
        "A string",
        np.random.RandomState(42),
    ],
)
def test_when_given_object_referenced_twice_loads_as_one_object(obj):
    an_object = {"obj_1": obj, "obj_2": obj}
    persisted_object = loads(dumps(an_object), trusted=True)

    assert persisted_object["obj_1"] is persisted_object["obj_2"]


class EstimatorWithBytes(BaseEstimator):
    def fit(self, X, y, **fit_params):
        self.bytes_ = b"hello"
        self.bytearray_ = bytearray([0, 1, 2, 253, 254, 255])
        return self


def test_estimator_with_bytes():
    est = EstimatorWithBytes().fit(None, None)
    loaded = loads(dumps(est), trusted=True)
    assert_params_equal(est.__dict__, loaded.__dict__)


def test_estimator_with_bytes_files_created(tmp_path):
    est = EstimatorWithBytes().fit(None, None)
    f_name = tmp_path / "estimator.skops"
    dump(est, f_name)
    file = Path(f_name)
    assert file.exists()

    with ZipFile(f_name, "r") as input_zip:
        files = input_zip.namelist()
    bin_files = [file for file in files if file.endswith(".bin")]
    assert len(bin_files) == 2


OPERATORS = [
    ("add", partial(operator.add, 0)),
    ("sub", partial(operator.sub, 0)),
    ("mul", partial(operator.mul, 1)),
    ("truediv", partial(operator.truediv, 1)),
    ("pow", partial(operator.pow, 1)),
    ("matmul", partial(operator.matmul, np.eye(N_SAMPLES))),
    ("iadd", partial(operator.iadd, 1)),
    ("isub", partial(operator.isub, 1)),
    ("imul", partial(operator.imul, 1)),
    ("itruediv", partial(operator.itruediv, 1)),
    ("ipow", partial(operator.ipow, 1)),
    # note: inplace matmul is not supported by numpy
    ("ge", partial(operator.ge, 0)),
    ("gt", partial(operator.gt, 0)),
    ("le", partial(operator.le, 0)),
    ("lt", partial(operator.lt, 0)),
    ("eq", partial(operator.eq, 0)),
    ("neg", operator.neg),
    ("attrgetter", operator.attrgetter("real")),
    ("attrgetter", operator.attrgetter("real", "real")),
    ("itemgetter", operator.itemgetter(None)),
    ("itemgetter", operator.itemgetter(None, None)),
    ("methodcaller", operator.methodcaller("round")),
    ("methodcaller", operator.methodcaller("round", 2)),
]

if sys.version_info >= (3, 11):
    OPERATORS.append(("call", partial(operator.call, len)))


@pytest.mark.parametrize("op", OPERATORS)
def test_persist_operator(op):
    # Test a couple of functions from the operator module. This is not an
    # exhaustive list but rather the most plausible functions. To check all
    # operators would require specific tests, not a generic one like this.
    # Fixes #283

    _, func = op
    # unfitted
    est = FunctionTransformer(func)
    loaded = loads(dumps(est), trusted=True)
    assert_params_equal(est.__dict__, loaded.__dict__)

    # fitted
    X, y = get_input(est)
    est.fit(X, y)
    loaded = loads(dumps(est), trusted=True)
    assert_params_equal(est.__dict__, loaded.__dict__)

    # Technically, we don't need to call transform. However, if this is skipped,
    # there is a danger to not define the function sufficiently, which could
    # lead to the test passing but being useless in practice -- e.g. for
    # func=operator.methodcaller, the test would pass without the fix for #283
    # but it would be useless in practice, since methodcaller is not properly
    # instantiated.
    est.transform(X)


@pytest.mark.parametrize("op", OPERATORS)
def test_persist_operator_raises_untrusted(op):
    # check that operators are not trusted by default, because at least some of
    # them could perform unsafe operations
    name, func = op
    est = FunctionTransformer(func)
    with pytest.raises(UntrustedTypesFoundException, match=name):
        loads(dumps(est), trusted=False)


def dummy_func(X):
    return X


@pytest.mark.parametrize("func", [np.sqrt, len, special.exp10, dummy_func])
def test_persist_function(func):
    estimator = FunctionTransformer(func=func)
    X, y = [0, 1], [2, 3]
    estimator.fit(X, y)

    dumped = dumps(estimator)
    loaded = loads(dumped, trusted=True)

    # check that loaded estimator is identical
    assert_params_equal(estimator.__dict__, loaded.__dict__)
    assert_method_outputs_equal(estimator, loaded, X)
